"""Mock framework for tests."""
import os
from typing import Generator
import xml.etree.ElementTree as ET
import pandas as pd
from tests import get_resources_path
from geniepy.datamgmt.scrapers import BaseScraper, PubtatorGeneScraper
from geniepy.pubmed import PubMedArticle
from geniepy.classmgmt import ClassificationMgr
from geniepy.classmgmt.classifiers import Classifier

TEST_CHUNKSIZE = 5
"""Default chunksize for tests."""


class MockClassifier(Classifier):
    """Mock classifier for tests."""

    def load(self):
        """
        Load classifier model.
        Raises:
            ClassifierError -- If model doesn't load successfully
        """
        self._model = True
        self._is_trained = True


PCPCLSFR = MockClassifier("pub_score")
CTCLSFR = MockClassifier("ct_score")
# pylint: disable=protected-access
CTCLSFR._col_name = "ct_score"
MOCK_CLSFRMGR: ClassificationMgr = ClassificationMgr([PCPCLSFR, CTCLSFR])


class MockPubtatorGeneScraper(PubtatorGeneScraper):
    def scrape(self, chunksize: int, **kwargs) -> Generator:
        """Mock scrape method."""
        header_names = ["PMID", "Type", "GeneID", "Mentions", "Resource"]
        vals = [[10001], ["Gene"], [112331], ["Mention"], ["Resources"]]
        zipped = list(zip(header_names, vals))
        yield pd.DataFrame(dict(zipped))


class MockPubMedScraper(BaseScraper):
    """Mock PubMed scraper for tests."""

    filename: str = "sample_articleset2.xml"

    def scrape(self, chunksize: int = 2, **kwargs) -> Generator:
        """Simulate scraping pubmed baseline and returning xml objs."""
        xml_file_path = os.path.join(get_resources_path(), self.filename)
        xml_root: ET.Element = ET.parse(xml_file_path).getroot()

        xml_list = xml_root.findall("PubmedArticle")
        pubmed_articles: [PubMedArticle] = []
        for article_xml in xml_list:
            pubmed_articles.append(PubMedArticle(article_xml))

        while True:
            articles_chunk = []
            for _ in range(chunksize):
                # Append articles to create chunk sized array
                if len(xml_list) > 0:
                    articles_chunk.append(xml_list.pop())
            # If no more articles return
            if not articles_chunk:
                return
            # If still articles, yield to generator
            yield articles_chunk


class MockCtdScraper(BaseScraper):
    """Mock CTD scraper for tests."""

    filename: str = "sample_ctd_db.csv"

    def scrape(self, chunksize: int, **kwargs) -> Generator:
        """Simulate scraping records and returning csv str."""
        csv_file = os.path.join(get_resources_path(), self.filename)
        # Open a connection to the file
        with open(csv_file) as file:
            # Read header
            header = file.readline()
            # Loop indefinitely until the end of the file
            while True:
                data = ""
                for _ in range(chunksize):
                    line = file.readline()
                    if not line:
                        break
                    data += line
                if not data:
                    break
                chunk = header + data
                yield chunk



"""Tests entry point module containing definition and helper functions."""

import os

RESOURCES_DIR = 'resources'
"""Name of test resources directory."""

OUTPUT_DIR = 'tests_output'
"""Name of temporary tests output directory to store test files."""


def get_resources_path() -> str:
    """Return absolute path to resources directory."""
    current_path = os.path.dirname(os.path.realpath(__file__))
    resources_path = os.path.join(current_path, RESOURCES_DIR)
    return resources_path


def gen_dir(dir_path: str):
    """Generate dir if doesn't exist."""
    if not os.path.exists(dir_path):
        os.mkdir(dir_path)


def get_test_output_path() -> str:
    """
    Get output path to test output directory.

    The test output directory is a temporary, automatically generated,
    directory to store test output and files.

    Returns:
        str -- absolute path to test output directory
    """
    current_path = os.path.dirname(os.path.relpath(__file__))
    output_path = os.path.join(current_path, OUTPUT_DIR)
    gen_dir(output_path)
    return output_path



# """Test Data Access Object Manager."""
# import pytest
# import geniepy.datamgmt.daos as daos
# import geniepy.datamgmt.repositories as dr
# from geniepy.datamgmt.tables import PUBMED_PROPTY, CTD_PROPTY, CLSFR_PROPTY
# from geniepy.datamgmt import DaoManager
# import tests.resources.mock as mock
# from tests.resources.mock import TEST_CHUNKSIZE


# class TestDaoManager:
#     """Pytest data access object manager test class."""

#     # Create and configure mock ctd dao
#     ctd_dao = daos.CtdDao(dr.SqlRepository("sqlite://", CTD_PROPTY))
#     # pylint: disable=protected-access
#     ctd_dao._parser.scraper = mock.MockCtdScraper()

#     # Create and configure mock pubmed dao
#     pubmed_dao = daos.PubMedDao(dr.SqlRepository("sqlite://", PUBMED_PROPTY))
#     # pylint: disable=protected-access
#     pubmed_dao._parser.scraper = mock.MockPubMedScraper()

#     # Create and configure mock classifier dao
#     classifier_dao = daos.ClassifierDao(dr.SqlRepository("sqlite://", CLSFR_PROPTY))
#     # pylint: disable=protected-access

#     # Construct mock dao manager for testing
#     dao_mgr = DaoManager(
#         ctd_dao=ctd_dao, pubmed_dao=pubmed_dao, classifier_dao=classifier_dao
#     )

#     def test_constructor(self):
#         """Test obj construction."""
#         assert self.dao_mgr is not None

#     def test_dowload_and_get_records(self):
#         """Test download and get_records functionality of DAO Mgr."""
#         gen_df = self.dao_mgr.gen_records(TEST_CHUNKSIZE)
#         # After instantiation dao's tables should be empty, so nothing is returned.
#         with pytest.raises(StopIteration):
#             next(gen_df)
#         # Call download function to scrape data and generate internal tables.
#         self.dao_mgr.download(TEST_CHUNKSIZE)
#         # Make sure gen_df is not none and conforms to expected schema
#         gen_df = self.dao_mgr.gen_records(TEST_CHUNKSIZE)
#         records = next(gen_df)
#         assert records is not None

#     @pytest.mark.parametrize("chunksize", [*range(1, 10)])
#     def test_get_records_chunksize(self, chunksize):
#         """Test get records chunksizes."""
#         self.dao_mgr.download(TEST_CHUNKSIZE)
#         gen_df = self.dao_mgr.gen_records(chunksize=chunksize)
#         records = next(gen_df)
#         # Make sure records contain correct chunk
#         assert records.shape[0] == chunksize

#     def test_save_predictions(self):
#         """Test writing predictions to output tables."""
#         self.dao_mgr.download(TEST_CHUNKSIZE)
#         gen_df = self.dao_mgr.gen_records(TEST_CHUNKSIZE)
#         records = next(gen_df)
#         predictions_df = mock.MOCK_CLSFRMGR.predict(records)
#         self.dao_mgr.save_predictions(predictions_df)
#         # Read data back from output tables to make sure records were saved
#         clsfr_dao = self.dao_mgr._classifier_dao
#         saved_df = next(clsfr_dao.query(clsfr_dao.query_all, TEST_CHUNKSIZE))
#         assert saved_df.equals(predictions_df)

# @pytest.mark.parametrize("chunksize", [*range(1, 10)])
# def test_get_records_chunksize(self, chunksize):
#     """Test get records chunksizes."""
#     self.dao_mgr.download(TEST_CHUNKSIZE)
#     gen_df = self.dao_mgr.gen_records(chunksize=chunksize)
#     records = next(gen_df)
#     # Make sure records contain correct chunk
#     assert records.shape[0] == chunksize



"""Module to test online sources parsers."""
import pytest
from geniepy.datamgmt.parsers import BaseParser, CtdParser
from geniepy.errors import ParserError
from tests.resources.mock import MockCtdScraper, TEST_CHUNKSIZE
import tests.testdata as td


VALID_DF = td.CTD_VALID_DF
INVALID_DF = td.CTD_INVALID_DF


class TestCtdParser:
    """Pytest CTD Parser class."""

    parser: BaseParser = CtdParser()
    mock_scraper = MockCtdScraper()
    parser.scraper = mock_scraper

    def test_constructor(self):
        """Ensure scraper obj constructed successfully."""
        assert self.parser is not None

    @pytest.mark.parametrize("payload", td.CTD_INVALID_DF)
    def test_invalid_payload(self, payload):
        """Test invalid dataframe."""
        # Should return list with errors
        assert self.parser.validate(payload)

    @pytest.mark.parametrize("payload", VALID_DF)
    def test_valid_payload(self, payload):
        """Test valid dataframe."""
        # Should return empty list
        assert not self.parser.validate(payload)

    def test_parse_valid(self):
        """Test parsing valid recrods."""
        mock_scraper = MockCtdScraper()
        scrape_gen = mock_scraper.scrape(TEST_CHUNKSIZE)
        self.parser.parse(next(scrape_gen))

    def test_parse_invalid_file(self):
        """Attempt to parse invalid data."""
        with pytest.raises(ParserError):
            self.parser.parse("invalid csv")

    def test_parse_invalid_df(self):
        """Attempt to parse invalid/incomplete dataframe."""
        mock_scraper = MockCtdScraper()
        mock_scraper.filename = "sample_corrupt_ctd_db.csv"
        scrape_gen = mock_scraper.scrape(TEST_CHUNKSIZE)
        with pytest.raises(ParserError):
            self.parser.parse(next(scrape_gen))

    def test_fetch(self):
        """Test parser fetch."""
        gen_data = self.parser.fetch(TEST_CHUNKSIZE)
        data = next(gen_data)
        assert data is not None



# """Pytest module to test Google BigQuery Repository."""
# import os
# import pytest
# import tests.testdata as td
# import geniepy.datamgmt.repositories as dr
# from geniepy.datamgmt.tables import PUBMED_PROPTY, CTD_PROPTY, CLSFR_PROPTY
# from geniepy.errors import DaoError
# from tests import get_test_output_path
# from tests.resources.mock import TEST_CHUNKSIZE


# # Name of credential file (assumed to be in tests/tests_output dir)
# CREDENTIALS_FILE = "genie_credentials.json"
# CREDPATH = os.path.join(get_test_output_path(), CREDENTIALS_FILE)
# # Google BigQuery Project Name
# PROJNAME = "genie-275215"
# DATASET = "test"


# VALID_DF = td.PUBMED_VALID_DF
# INVALID_SCHEMA = td.PUBMED_INVALID_SCHEMA


# @pytest.mark.slow_integration_test
# class TestGbqRepository:
#     """Test db.pubmed_repository on Google BigQuery."""

#     pubmed_repo: dr.BaseRepository = None
#     ctd_repo: dr.BaseRepository = None
#     clsfr_repo: dr.BaseRepository = None

#     @classmethod
#     def setup_class(cls):
#         """Initialize GBQ.pubmed_repo."""
#         cls.pubmed_repo = dr.GbqRepository(PROJNAME, PUBMED_PROPTY, DATASET, CREDPATH)
#         cls.ctd_repo = dr.GbqRepository(PROJNAME, CTD_PROPTY, DATASET, CREDPATH)
#         cls.clsfr_repo = dr.GbqRepository(PROJNAME, CLSFR_PROPTY, DATASET, CREDPATH)

#     def test_constructor(self):
#         """Test constructing object."""
#         assert self.pubmed_repo is not None

#     @pytest.mark.parametrize("payload", INVALID_SCHEMA)
#     def test_save_invalid_df(self, payload):
#         """Test save invalid dataframe to dao's DAO."""
#         with pytest.raises(DaoError):
#             self.pubmed_repo.save(payload)

#     @pytest.mark.parametrize("payload", VALID_DF)
#     def test_save_valid_df(self, payload):
#         """Attempt to save dataframe with valid schema."""
#         self.pubmed_repo.save(payload)  # Don't expect to return anything

#     def test_pubmed_query(self):
#         """Query valid record."""
#         payload = VALID_DF[0]
#         # Start with empty table
#         self.pubmed_repo.delete_all()
#         # Try to create records in db for test if don't exist
#         try:
#             self.pubmed_repo.save(payload)
#         except DaoError:
#             pass
#         # Attempt to retrieve record
#         pmid = payload.pmid[0]
#         query_str = self.pubmed_repo.query_pkey(pmid)
#         generator = self.pubmed_repo.query(query_str, TEST_CHUNKSIZE)
#         chunk = next(generator)
#         assert chunk.pmid.equals(payload.pmid)

#     def test_ctd_query(self):
#         """Query valid record."""
#         payload = td.CTD_VALID_DF[0]
#         # Start with empty table
#         self.ctd_repo.delete_all()
#         # Try to create records in db for test if don't exist
#         try:
#             self.ctd_repo.save(payload)
#         except DaoError:
#             pass
#         # Attempt to retrieve record
#         digest = payload.digest[0]
#         query_str = self.ctd_repo.query_pkey(digest)
#         generator = self.ctd_repo.query(query_str, TEST_CHUNKSIZE)
#         chunk = next(generator)
#         assert chunk.digest.equals(payload.digest)

#     def test_clsfr_query(self):
#         """Query valid record."""
#         payload = td.CLSFR_VALID_DF[0]
#         # Start with empty table
#         self.clsfr_repo.delete_all()
#         # Try to create records in db for test if don't exist
#         try:
#             self.clsfr_repo.save(payload)
#         except DaoError:
#             pass
#         # Attempt to retrieve record
#         digest = payload.digest[0]
#         query_str = self.clsfr_repo.query_pkey(digest)
#         generator = self.clsfr_repo.query(query_str, TEST_CHUNKSIZE)
#         chunk = next(generator)
#         assert chunk.digest.equals(payload.digest)

#     def test_invalid_query(self):
#         """Test making invalid queries."""
#         query_str = "Invalid"
#         with pytest.raises(DaoError):
#             next(self.pubmed_repo.query(query_str, TEST_CHUNKSIZE))

#     def test_query_non_existent(self):
#         """Query non-existent record should return empty."""
#         # Attempt to retrieve record
#         pmid = 0
#         query_str = self.pubmed_repo.query_pkey(pmid)
#         generator = self.pubmed_repo.query(query_str, TEST_CHUNKSIZE)
#         # Make sure generator doesn't return anything since no matching records
#         with pytest.raises(StopIteration):
#             next(generator)

#     @pytest.mark.parametrize("chunksize", [1, 2, 3])
#     def test_generator_chunk(self, chunksize):
#         """Query all by chunk."""
#         # Start with empty table
#         self.pubmed_repo.delete_all()
#         # Try to fill database, in case is empty
#         for record in VALID_DF:
#             try:
#                 self.pubmed_repo.save(record)
#             except DaoError:
#                 pass
#         # Get all records in database
#         query_all = self.pubmed_repo.query_all
#         generator = self.pubmed_repo.query(query_all, chunksize)
#         # Make sure number generator provides df of chunksize each iteration
#         result_df = next(generator)
#         assert result_df.pmid.count() == chunksize

#     def test_delete_all(self):
#         """Test delete all records from.pubmed_repository."""
#         # Try to fill database, in case is empty
#         for record in VALID_DF:
#             try:
#                 self.pubmed_repo.save(record)
#             except DaoError:
#                 pass
#         # Delete all records
#         self.pubmed_repo.delete_all()
#         # Make sure no records left
#         query_all = self.pubmed_repo.query_all
#         generator = self.pubmed_repo.query(query_all, TEST_CHUNKSIZE)
#         # generator shouldn't return anything since no records in database
#         with pytest.raises(StopIteration):
#             next(generator)
#         # Test saving and reading from table again, make sure still functional
#         self.pubmed_repo.save(VALID_DF[0])
#         generator = self.pubmed_repo.query(query_all, TEST_CHUNKSIZE)
#         # Generator should return value
#         next(generator)



"""Module to test Data Access Objects."""
import pytest
from geniepy.datamgmt.daos import BaseDao, ClassifierDao
from geniepy.errors import SchemaError
import geniepy.datamgmt.repositories as dr
from geniepy.datamgmt.tables import CLSFR_PROPTY
from geniepy.errors import DaoError
import tests.testdata as td
from tests.resources.mock import TEST_CHUNKSIZE

VALID_DF = td.CLSFR_VALID_DF
INVALID_DF = td.CLSFR_INVALID_DF


class TestClassifierDao:
    """PyTest data access object test class."""

    test_repo = dr.SqlRepository("sqlite://", CLSFR_PROPTY)
    test_dao: BaseDao = ClassifierDao(test_repo)

    def read_record(self, digest):
        """Read record(s) from database (tests helper method)."""
        query_str = self.test_dao.query_pkey(digest)
        generator = self.test_dao.query(query_str, TEST_CHUNKSIZE)
        return generator

    def test_constructor(self):
        """Ensure obj constructed successfully."""
        assert self.test_dao is not None

    @pytest.mark.parametrize("payload", INVALID_DF)
    def test_save_invalid_df(self, payload):
        """Test save invalid dataframe to dao's repository."""
        with pytest.raises(SchemaError):
            self.test_dao.save(payload)

    @pytest.mark.parametrize("payload", VALID_DF)
    def test_save_valid_df(self, payload):
        """Test save valid dataframe to dao's repo doesn't raise error."""
        self.test_dao.save(payload)

    @pytest.mark.parametrize("payload", VALID_DF)
    def test_query(self, payload):
        """Query valid record."""
        # Start with empty table
        self.test_dao.purge()
        # Try to create records in db for test if don't exist
        try:
            self.test_dao.save(payload)
        except DaoError:
            pass
        # Attempt to retrieve record
        digest = payload.digest[0]
        generator = self.read_record(digest)
        chunk = next(generator)
        assert chunk.equals(payload)

    def test_query_non_existent(self):
        """Query non-existent record should return empty."""
        # Attempt to retrieve record
        digest = "INVALID digest"
        generator = self.read_record(digest)
        # Make sure generator doesn't return anything since no records in database
        with pytest.raises(StopIteration):
            next(generator)

    def test_purge(self):
        """Test delete all records from repository."""
        # Try to fill database, in case is empty
        for record in VALID_DF:
            try:
                self.test_dao.save(record)
            except DaoError:
                pass
        # Delete all records
        self.test_dao.purge()
        # Make sure no records left
        generator = self.test_dao.query(self.test_dao.query_all, TEST_CHUNKSIZE)
        # generator shouldn't return anything since no records in database
        with pytest.raises(StopIteration):
            next(generator)
        # Test building and reading from table again, make sure still functional
        self.test_query(VALID_DF[0])

    @pytest.mark.parametrize("chunksize", [*range(1, len(VALID_DF) + 1)])
    def test_query_chunksize(self, chunksize):
        """Query all by chunk."""
        # Try to fill database, in case is empty
        for record in VALID_DF:
            try:
                self.test_dao.save(record)
            except DaoError:
                pass
        # Get all records in database
        generator = self.test_dao.query(self.test_dao.query_all, chunksize)
        # Make sure number generator provides df of chunksize each iteration
        result_df = next(generator)
        assert result_df.digest.count() == chunksize

    def test_download_not_impl(self):
        """Download method not impl in classifier dao."""
        with pytest.raises(NotImplementedError):
            self.test_dao.download(TEST_CHUNKSIZE)



"""Data sources parsers."""
from typing import Generator
import hashlib
from abc import ABC, abstractstaticmethod
from enum import Enum, auto
from io import StringIO
import numpy as np
import pandas as pd
from pandas import DataFrame
from pandas_schema import Column, Schema
from pandas_schema.validation import IsDtypeValidation, MatchesPatternValidation
import geniepy.datamgmt.scrapers as gs
from geniepy.errors import ParserError
from geniepy.pubmed import PubMedArticle
from geniepy.classmgmt.classifiers import PCPCLSFR_NAME, CTCLSFR_NAME


class DataType(Enum):
    """Possible parsable datatypes."""

    CSV_STR = auto()  # csv string
    XML = auto()  # xml element tree
    DF = auto()  # pandas dataframe


class BaseParser(ABC):
    """Abstract base parser class."""

    scraper: gs.BaseScraper
    schema: Schema
    default_type: DataType = None

    @classmethod
    def validate(cls, payload: DataFrame) -> [str]:
        """
        Check if payload is valid schema.

        Arguments:
            payload {DataFrame} -- The data to be checked against parser schema.

        Returns:
            bool -- true if payload conforms to schema, false otherwise.
        """
        if payload is None:
            return ["Cannot validate None object"]
        return cls.schema.validate(payload)

    @abstractstaticmethod
    def parse(data, dtype: DataType = None) -> DataFrame:
        """
        Parse data and convert according to parser schema.

        Arguments:
            data {Implementation dependent} -- Data to be parsed

        Keyword Arguments:
            dtype {DataType} -- Type of data to be parsed (default: {DataType.CSV})

        Returns:
            DataFrame -- The parsed dataframe.
        """

    def fetch(self, chunksize: int, **kwargs) -> Generator[DataFrame, None, None]:
        """
        Fetch new data, if available from online sources.

        Keyword Arguments:
            chunksize {int} -- the returned generator chunk size

        Returns:
            Generator[DataFrame, None, None] -- Generator yielding fetched data
        """
        raw_gen = self.scraper.scrape(chunksize, **kwargs)

        # for testing purposes only - returns single yield
        if "is_sample" in kwargs and kwargs.get("is_sample") == True:
            parsed_df = self.parse(next(raw_gen), self.default_type)
            yield parsed_df
            return

        # for production purposes - iterate through yields
        for data_chunk in raw_gen:
            parsed_df = self.parse(data_chunk, self.default_type)
            yield parsed_df


class CtdParser(BaseParser):
    """
    Implementation of CTD Database Parser.

    Comparative Toxicogenomics Gene-Disease Associations Database Parser.
    http://ctdbase.org/
    """

    default_type: DataType = DataType.CSV_STR
    scraper: gs.CtdScraper = gs.CtdScraper()
    schema: Schema = Schema(
        [
            Column("digest"),
            Column("genesymbol"),
            Column("geneid", [IsDtypeValidation(np.int64)]),
            Column("diseasename"),
            Column(
                "diseaseid", [MatchesPatternValidation("^D[0-9]+$")]
            ),  # i.e. D000014
            Column("pmids"),
        ]
    )

    @staticmethod
    def hash_record(record: pd.Series) -> str:
        """
        Hash the ctd record to generate digest column.

        Arguments:
            record {pd.Series} -- The ctd record in form of pandas Series

        Returns:
            str -- the hex string of the computed digest
        """
        message = str.encode(str(record.geneid) + record.diseaseid)
        hexdigest = hashlib.sha256(message).hexdigest()
        return str(hexdigest)

    @staticmethod
    def parse(data, dtype=DataType.CSV_STR) -> DataFrame:
        """
        Parse data and convert according to parser schema.

        Arguments:
            data {Implementation dependent} -- Data to be parsed

        Keyword Arguments:
            dtype {DataType} -- Type of data to be parsed (default: {DataType.CSV})

        Returns:
            DataFrame -- The parsed dataframe.

        Raises:
            ParserError -- If unable to parse data
        """
        try:
            parsed_df = pd.read_csv(StringIO(data))
            # Remove unused columns
            parsed_df = parsed_df.drop(
                columns=[
                    "DirectEvidence",
                    "InferenceChemicalName",
                    "InferenceScore",
                    "OmimIDs",
                ]
            )
            # Remove prefix 'MESH:' from DiseaseIDs
            parsed_df["DiseaseID"] = parsed_df.apply(
                lambda x: x.DiseaseID.replace("MESH:", ""), axis=1
            )
            # Rename columns based on schema
            parsed_df.rename(
                columns={
                    "GeneSymbol": "genesymbol",
                    "GeneID": "geneid",
                    "DiseaseName": "diseasename",
                    "DiseaseID": "diseaseid",
                    "PubMedIDs": "pmids",
                },
                inplace=True,
            )
            # Compute and add the digest
            parsed_df["digest"] = parsed_df.apply(CtdParser.hash_record, axis=1)
            errors = CtdParser.validate(parsed_df)
            if errors:
                raise ParserError(errors)
            return parsed_df
        except Exception as parse_exp:
            raise ParserError(parse_exp)


class PubtatorDiseaseParser(BaseParser):
    """Implementation of Pubtator Disease Parser."""

    default_type: DataType = DataType.DF
    scraper: gs.PubtatorDiseaseScraper = gs.PubtatorDiseaseScraper()

    @staticmethod
    def parse(data, dtype=DataType.DF) -> DataFrame:
        """Parse data and convert according to parser schema."""
        try:
            parsed_df = data[["PMID", "DiseaseID"]].copy()
            return parsed_df
        except:
            return None


class PubtatorGeneParser(BaseParser):
    """Implementation of Pubtator Gene Parser."""

    default_type: DataType = DataType.DF
    scraper: gs.PubtatorGeneScraper = gs.PubtatorGeneScraper()

    @staticmethod
    def parse(data, dtype=DataType.DF) -> DataFrame:
        """Parse data and convert according to parser schema."""
        try:
            parsed_df = data[["PMID", "GeneID"]].copy()
            return parsed_df
        except:
            return None


class SjrParser(BaseParser):
    """Implementation of Scientific Journal Ratings Parser."""

    default_type: DataType = DataType.DF
    scraper: gs.SjrScraper = gs.SjrScraper()

    @staticmethod
    def parse(data, dtype=DataType.DF) -> DataFrame:
        """Parse data and convert according to parser schema."""
        try:
            parsed_df = data[["Title", "SJR", "H index"]].copy()
            parsed_df.rename(columns={"H index": "h_index"}, inplace=True)
            return parsed_df
        except:
            return None


class PubMedParser(BaseParser):
    """
    Implementation of PubMed Articles Parser.

    https://www.ncbi.nlm.nih.gov/pubmed/
    https://www.nlm.nih.gov/bsd/licensee/elements_descriptions.html
    """

    default_type: DataType = DataType.XML
    scraper: gs.PubMedScraper = gs.PubMedScraper()
    schema: Schema = Schema(
        [
            Column("pmid", [IsDtypeValidation(np.int64)]),
            Column("date_completed"),
            Column("pub_model"),
            Column("title"),
            Column("iso_abbreviation"),
            Column("article_title"),
            Column("abstract"),
            Column("authors"),
            Column("language"),
            Column("chemicals"),
            Column("mesh_list"),
            Column("issn"),
            Column("issn_type"),
            Column("citation_count"),
            Column("citation_pmid"),
        ]
    )

    @staticmethod
    def parse(data, dtype: DataType = None) -> DataFrame:
        """
        Parse data and convert according to parser schema.

        Arguments:
            data {Implementation dependent} -- Data to be parsed

        Keyword Arguments:
            dtype {DataType} -- Type of data to be parsed (default: {DataType.CSV})

        Returns:
            DataFrame -- The parsed dataframe.
        """
        # Data passed in should be a list of xml element trees
        # xml_list = data
        articles = data

        # The keys of the dataframe
        keys = [
            "pmid",
            "date_completed",
            "pub_model",
            "title",
            "iso_abbreviation",
            "article_title",
            "abstract",
            "authors",
            "language",
            "chemicals",
            "mesh_list",
            "issn",
            "issn_type",
            "citation_count",
            "citation_pmid",
        ]

        # Temp array variables to store from each xml element tree
        pmid_list = []
        date_completed_list = []
        pub_model_list = []
        title_list = []
        iso_abbreviation_list = []
        article_title_list = []
        abstract_list = []
        authors_list = []
        language_list = []
        chemicals_list = []
        mesh_list_list = []
        issn_list = []
        issn_type_list = []
        citation_count_list = []
        citation_pmid_list = []

        try:
            # General XML Tags
            # for xml_article in xml_list:
            for article in articles:
                # article = PubMedArticle(xml_article)
                pmid_list.append(np.int64(article.pmid))
                date_completed_list.append(article.date_completed)
                pub_model_list.append(article.pub_model)
                title_list.append(article.title)
                iso_abbreviation_list.append(article.iso_abbreviation)
                article_title_list.append(article.article_title)
                abstract_list.append(article.abstract)
                authors_list.append(str(article.authors).strip("[]"))
                language_list.append(article.language)
                chemicals_list.append(str(article.chemicals).strip("[]"))
                mesh_list_list.append(str(article.mesh_list).strip("[]"))
                issn_list.append(str(article.issn).strip("[]"))
                issn_type_list.append(str(article.issn_type).strip("[]"))
                citation_count_list.append(str(article.citationCount).strip("[]"))
                citation_pmid_list.append(str(article.citationPmid).strip("[]"))

            # Create the array of arrays of values in dataframe
            values = [
                pmid_list,
                date_completed_list,
                pub_model_list,
                title_list,
                iso_abbreviation_list,
                article_title_list,
                abstract_list,
                authors_list,
                language_list,
                chemicals_list,
                mesh_list_list,
                issn_list,
                issn_type_list,
                citation_count_list,
                citation_pmid_list,
            ]

            # Zip df keys and values and create dataframe
            zipped = list(zip(keys, values))
            parsed_df = pd.DataFrame(dict(zipped))

            errors = PubMedParser.validate(parsed_df)
            if errors:
                raise ParserError(errors)  # pragma: no cover - should never reach
            return parsed_df
        except Exception as parse_exp:
            raise ParserError(parse_exp)


class ClassifierParser(BaseParser):
    """
    Implementation of classifier dao Parser.

    The classifier output tables contain the output data from geniepy after the
    classifiers have calculated desired predictions.
    """

    default_type: DataType = None
    scraper: None
    """No online sources for classifiers output."""
    schema: Schema = Schema(
        [
            Column("digest"),
            Column(PCPCLSFR_NAME, [IsDtypeValidation(np.float64)]),
            Column(CTCLSFR_NAME, [IsDtypeValidation(np.float64)]),
        ]
    )

    def fetch(self, chunksize: int) -> Generator[DataFrame, None, None]:
        """No online sources to fetch from for classifiers outputs."""
        raise NotImplementedError("Classifier Output Parser has no Scrapers")

    @staticmethod
    def parse(data, dtype=DataType.CSV_STR) -> DataFrame:
        """
        Parser function from base class.

        Raises:
            NotImplementedError -- Function not implemented since classifiers return
                dataframes that only need to be validated.
        """
        raise NotImplementedError("Classifier Output Parser has no Scrapers")



"""
This script is deployed as Google Cloud Function.

It allows to scale-up the citation metadata extraction process.
This script also utilizes multiple PubMed API Keys, to scale-up the number of API
calls that can be made per second. Plesae make sure the API Keys included are
valid before executing this script.
"""
from flask import abort
import json
from random import randint
import requests
import xml.etree.ElementTree as ET

# Constants
API_BASE_URL = "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/elink.fcgi?dbfrom=pubmed&linkname=pubmed_pubmed_citedin"  # noqa
API_ID_PARAM = "&id="
API_KEY_PARAM = "&api_key="
API_KEY_1 = "70b53d4e84436970587aef3493a723cae708"
API_KEY_2 = "9ab4b04dcabcab740d1a297e6f3f54aa1e09"
API_KEY_3 = "39458b83e654b32523475ee6b828b6c22c08"
API_KEY_4 = "cf95d471b41c4ea9e6b89d41de269705eb08"
API_KEY_5 = "94c3049a7b3dffca2d22ce11b51a3ea5da09"
API_KEY_6 = "3fb37654e8f711f8751d9e3c21fc46257108"
API_KEY_7 = "91315e4e69c1b5e2676371490cfc715caa08"
API_KEY_8 = "b1ee76fae01c312e4bf2fd3e8b1bcf978a09"
API_KEY_9 = "25b74bb456d7e6fc756c86c7aebb1b525009"
API_KEY_10 = "7a7e4492084dd09cef758ee65a7909704b08"
API_KEY_11 = "4d32405f13fed69766ad639d89c5c5288608"
API_KEY_12 = "18c8d26e60b5cd012e2ba9618c88dd12cb09"
API_KEY_13 = "667ca5b1b72b7697ed43c26727329282bb08"
API_KEY_14 = "5cdea4b682e2fc93c56c2b1e5e9da93f4709"
API_KEY_15 = "108b2e1a81369991da6c737545097c581a08"
API_KEY_16 = "f10805158a7609bb115cb074b05e5923a407"
API_KEY_17 = "00d74dd1cb732cd7fc29f54168fe7055c809"
TAG_LINK_SET = "LinkSet"
TAG_ID = "./IdList/Id"
TAG_CITATION_ID = "./LinkSetDb/Link/Id"


def get_random_api_key():  # noqa
    # pick a random number between 1 and 5
    num = randint(1, 17)
    # return a randomly selected cloud function
    if num == 1:
        return API_KEY_1
    elif num == 2:
        return API_KEY_2
    elif num == 3:
        return API_KEY_3
    elif num == 4:
        return API_KEY_4
    elif num == 5:
        return API_KEY_5
    elif num == 6:
        return API_KEY_6
    elif num == 7:
        return API_KEY_7
    elif num == 8:
        return API_KEY_8
    elif num == 9:
        return API_KEY_9
    elif num == 10:
        return API_KEY_10
    elif num == 11:
        return API_KEY_11
    elif num == 12:
        return API_KEY_12
    elif num == 13:
        return API_KEY_13
    elif num == 14:
        return API_KEY_14
    elif num == 15:
        return API_KEY_15
    elif num == 16:
        return API_KEY_16
    else:
        return API_KEY_17


def parse_citation(url: str):

    _citation_dict = {}

    try:
        req = requests.get(url)
        tree = ET.fromstring(req.text)
    except Exception:
        print(f"Exeception: bad request: {url}")
        return _citation_dict

    for linkSet in tree.findall(TAG_LINK_SET):
        _pubmed_id = linkSet.find(TAG_ID).text
        _citation_id_list: str = []
        for citationId in linkSet.findall(TAG_CITATION_ID):
            _citation_id_list.append(citationId.text)

        if not _citation_id_list or len(_citation_id_list) <= 0:
            _citation_id_list.append("0")

        _citation_dict[_pubmed_id] = _citation_id_list

    return _citation_dict


def get_citation(request):
    if request.method != "POST":
        return abort(405)

    # get list of PubMed IDs from the request body
    request_json = request.get_json()
    pubmed_ids = request_json["pubmed_ids"]

    # split PubMed IDs and construct ID Parameter list
    API_ID_PARAM_LIST = ""
    for pubmed_id in pubmed_ids.split(","):
        API_ID_PARAM_LIST = API_ID_PARAM_LIST + API_ID_PARAM + pubmed_id

    # get random API Key
    API_KEY = get_random_api_key()

    # construct API url
    url = API_BASE_URL + API_KEY_PARAM + API_KEY + API_ID_PARAM_LIST

    return json.dumps(parse_citation(url))



"""Test pub med article model class."""
import os
import collections
import xml.etree.ElementTree as ET
import pytest
from tests import get_resources_path, get_test_output_path
from geniepy.pubmed import PubMedArticle, ArticleSetParser


def create_article(article_name: str) -> PubMedArticle:
    """Create the article object."""
    article_path = os.path.join(get_resources_path(), article_name)
    xml_element = ET.parse(article_path).getroot()
    return PubMedArticle(xml_element)


ExpectedArticle = collections.namedtuple(
    "ExpectedArticle",
    "pmid date_completed pub_model title iso_abbrev \
                        article_title abstract authors language chemicals \
                        mesh_list",
)
ActualExpected = collections.namedtuple("TestData", "article expected")
TEST_DATA = [
    ActualExpected(
        article=create_article("sample_empty_article.xml"),
        expected=ExpectedArticle(
            pmid="",
            date_completed="--",
            pub_model="",
            title="",
            iso_abbrev="",
            article_title="",
            abstract="",
            authors=[],
            language="",
            chemicals=[],
            mesh_list=[],
        ),
    ),
    ActualExpected(
        article=create_article("sample_article1.xml"),
        expected=ExpectedArticle(
            pmid="1",
            date_completed="1976-01-16",
            pub_model="Print",
            title="Biochemical medicine",
            iso_abbrev="Biochem Med",
            article_title="Formate assay in body fluids: \
application in methanol poisoning.",
            abstract="",
            authors=["Makar, A B", "McMartin, K E", "Palese, M", "Tephly, T R"],
            language="eng",
            chemicals=[
                "Formates",
                "Carbon Dioxide",
                "Methanol",
                "Aldehyde Oxidoreductases",
            ],
            mesh_list=[],
        ),
    ),
    ActualExpected(
        article=create_article("sample_article2.xml"),
        expected=ExpectedArticle(
            pmid="15299998",
            date_completed="2006-02-03",
            pub_model="Print",
            title="Medicina (Kaunas, Lithuania)",
            iso_abbrev="Medicina (Kaunas)",
            article_title="[The crop-producing power and chemical \
composition of the essential oil of the cones of hop cultivars].",
            abstract='The objective of this study was to determine \
harvest yield and essential oil composition of six hop cultivars \
("Fredos derlingieji", "Alta", "Granit", "Marynka", "Dubskij zeleniak" \
and "Aromat polessja") cultivated in Lithuania. It was found that the highest \
crop can be obtained from early and middle harvest varieties, which mature \
during the first decade of September and the third decade of August. Late \
harvest cultivars are not suitable for cultivation, however they can be used \
for the development of new cultivars. Essential oils were isolated by \
hydrodistillation and analyzed by gas chromatography and mass \
spectrometry. The highest amount of essential oil was determined \
in Marynka cultivar (2.10 ml/100 g), the lowest one in Dubskij \
zeleniak 18 (0.46 ml/100 g). Totally, 62 compounds were identified in all \
cultivars. Myrcene, beta-caryophyllene, alpha-humulene and beta-farnesene \
(E) were major constituents in the essential oils. The differences in the \
content of other essential oil constituents were determined. Some cultivars \
were rich in esters and ketones.',
            authors=["Obelevicius, Kestutis", "Venskutonis, Rimantas"],
            language="lit",
            chemicals=[
                "Monocyclic Sesquiterpenes",
                "Oils, Volatile",
                "Polycyclic Sesquiterpenes",
                "Sesquiterpenes",
                "beta-farnesene",
                "humulene",
                "caryophyllene",
            ],
            mesh_list=[
                "Chromatography, Gas",
                "Climate",
                "Humans",
                "Humulus",
                "Lithuania",
                "Mass Spectrometry",
                "Monocyclic Sesquiterpenes",
                "Oils, Volatile",
                "Polycyclic Sesquiterpenes",
                "Seasons",
                "Sesquiterpenes",
            ],
        ),
    ),
]


class TestPubMedArticle:
    """Test class for functionlities related to PubMedArticle class."""

    @pytest.mark.parametrize("data", TEST_DATA)
    def test_article_constructor(self, data):
        """Construct pubmed article object from xml."""
        assert data.article is not None

    @pytest.mark.parametrize("data", TEST_DATA)
    def test_pmid(self, data):
        """Verify article id."""
        expected_pmid = data.expected.pmid
        pmid = data.article.pmid
        assert pmid == expected_pmid

    @pytest.mark.parametrize("data", TEST_DATA)
    def test_date_completed(self, data):
        """Verify article date completed."""
        expected_date = data.expected.date_completed
        date_completed = data.article.date_completed
        assert date_completed == expected_date

    @pytest.mark.parametrize("data", TEST_DATA)
    def test_article_pubmodel(self, data):
        """Verify PubModel - medium which cited article was published."""
        expected_pub_model = data.expected.pub_model
        pub_model = data.article.pub_model
        assert pub_model == expected_pub_model

    @pytest.mark.parametrize("data", TEST_DATA)
    def test_title(self, data):
        """Verify title of journal."""
        expected_title = data.expected.title
        title = data.article.title
        assert title == expected_title

    @pytest.mark.parametrize("data", TEST_DATA)
    def test_iso_abbrev(self, data):
        """The title ISO abbreviation."""
        expected_iso_abbrev = data.expected.iso_abbrev
        iso_abbrev = data.article.iso_abbreviation
        assert iso_abbrev == expected_iso_abbrev

    @pytest.mark.parametrize("data", TEST_DATA)
    def test_article_title(self, data):
        """Entire title of journal article in English."""
        expected_article_title = data.expected.article_title
        article_title = data.article.article_title
        assert article_title == expected_article_title

    @pytest.mark.parametrize("data", TEST_DATA)
    def test_abstract(self, data):
        """Verify abstract of published article."""
        expected_abstract = data.expected.abstract
        abstract = data.article.abstract
        assert abstract == expected_abstract

    @pytest.mark.parametrize("data", TEST_DATA)
    def test_authors(self, data):
        """Verify names of authors."""
        expected_set = set(data.expected.authors)
        authors_set = set(data.article.authors)
        assert authors_set == expected_set

    @pytest.mark.parametrize("data", TEST_DATA)
    def test_language(self, data):
        """Verify article language."""
        expected_language = data.expected.language
        language = data.article.language
        assert language == expected_language

    @pytest.mark.parametrize("data", TEST_DATA)
    def test_chemical_list(self, data):
        """Verify list of chemicals."""
        expected_set = set(data.expected.chemicals)
        chemicals_set = set(data.article.chemicals)
        assert chemicals_set == expected_set

    @pytest.mark.parametrize("data", TEST_DATA)
    def test_mesh_list(self, data):
        """Verify mesh descriptors list."""
        expected_set = set(data.expected.mesh_list)
        mesh_set = set(data.article.mesh_list)
        assert mesh_set == expected_set

    @pytest.mark.parametrize("data", TEST_DATA)
    def test_dict(self, data):
        """Test object dictionary."""
        assert data.article.to_dict is not None


class TestArticlesSetParser:
    """Test ArticleSetParser. i.e. xml files with array of PubMedArticles."""

    SAMPLE_ARTICLE_SET1_NAME = "sample_articleset1.xml"
    SAMPLE_ARTICLE_SET1_PATH = os.path.join(
        get_resources_path(), SAMPLE_ARTICLE_SET1_NAME
    )

    def test_extract_articles(self):
        """Extract pubmed articles from article set."""
        articles: [PubMedArticle] = ArticleSetParser.extract_articles(
            self.SAMPLE_ARTICLE_SET1_PATH
        )
        assert articles is not None
        assert len(articles) == 2
        assert isinstance(articles[0], PubMedArticle)

    def test_articles_to_json(self):
        """Serialize pubmedarticle object into json file."""
        articles: [PubMedArticle] = ArticleSetParser.extract_articles(
            self.SAMPLE_ARTICLE_SET1_PATH
        )
        target_file_name = "test_articles.json"
        target_path = os.path.join(get_test_output_path(), target_file_name)
        ArticleSetParser.articles_to_json(articles, target_path)

    def test_articles_to_jsonl(self):
        """Serialize pubmedarticles objects into jsonl file."""
        articles: [PubMedArticle] = ArticleSetParser.extract_articles(
            self.SAMPLE_ARTICLE_SET1_PATH
        )
        target_file_name = "test_articles.jsonl"
        target_path = os.path.join(get_test_output_path(), target_file_name)
        ArticleSetParser.articles_to_jsonl(articles, target_path)

    def test_serialize_to_pipe_delimited(self):
        """Serialize pubmedarticles to csv file."""
        articles: [PubMedArticle] = ArticleSetParser.extract_articles(
            self.SAMPLE_ARTICLE_SET1_PATH
        )
        target_file_name = "test_articles.csv"
        target_path = os.path.join(get_test_output_path(), target_file_name)
        ArticleSetParser.articles_to_pipe(articles, target_path)



"""GeniePy Package entry point."""
import os
import sys
import warnings
from concurrent.futures import ProcessPoolExecutor
from itertools import repeat
from pkg_resources import get_distribution, DistributionNotFound
from time import time
import geniepy.config as config
from geniepy.datamgmt import DaoManager
import geniepy.datamgmt.daos as daos
import geniepy.datamgmt.repositories as dr
import geniepy.datamgmt.tables as gt
from geniepy.classmgmt import ClassificationMgr
from geniepy.classmgmt.classifiers import Classifier
from geniepy.classmgmt.classifiers import PCPCLSFR_NAME, CTCLSFR_NAME
from geniepy.errors import ConfigError


try:
    # Change here if project is renamed and does not equal the package name
    DIST_NAME = __name__
    __version__ = get_distribution(DIST_NAME).version
except DistributionNotFound:
    __version__ = "unknown"
finally:
    del get_distribution, DistributionNotFound

__author__ = "The Harvard LAMP Team"
__copyright__ = "The Harvard LAMP Team"
__license__ = "MIT"

__all__ = ["run", "run_predictions", "update_tables"]


def create_repos():
    """Configure and create scoring repositories."""
    try:
        credentials = config.get_credentials()
        projname = config.get_projname()
        dataset = config.get_dataset("scoring")
        features_repo = dr.GbqRepository(
            projname, gt.FEATURES_PROPTY, dataset, credentials
        )
        scores_repo = dr.GbqRepository(projname, gt.SCORES_PROPTY, dataset, credentials)
        return features_repo, scores_repo
    except ConfigError as exp:
        print(exp.message)
        exit()


def create_dao(daoname: str):
    """Configure and create Data Access Object."""
    dao_dict = {
        "disease2pubtator": (daos.PubtatorDiseaseDao, gt.PUBTATOR_DISEASE_PROPTY,),
        "gene2pubtator": (daos.PubtatorGeneDao, gt.PUBTATOR_GENE_PROPTY),
        "sjr": (daos.SjrDao, gt.SJR_PROPTY),
        "pubmed": (daos.PubMedDao, gt.PUBMED_PROPTY),
    }
    DAO_CLS = dao_dict[daoname][0]
    TABLE_PROPTY = dao_dict[daoname][1]
    credentials = config.get_credentials()
    projname = config.get_projname()
    dataset = config.get_dataset("master")
    dao = DAO_CLS(dr.GbqRepository(projname, TABLE_PROPTY, dataset, credentials))
    return dao


def create_daomgr() -> DaoManager:
    """Configure data mgmt subsystem."""
    try:
        sjr_dao = create_dao("sjr")
        pubtator_disease_dao = create_dao("disease2pubtator")
        pubtator_gene_dao = create_dao("gene2pubtator")
        pubmed_dao = create_dao("pubmed")
        daomgr = DaoManager(
            sjr_dao, pubtator_disease_dao, pubtator_gene_dao, pubmed_dao
        )
        return daomgr
    except Exception:
        print("Unable to configure Dao Manager")


def create_classmgr() -> ClassificationMgr:
    """Configure classification mgmt subsystem."""
    # TODO Retrieve from config file
    pub_clsfr = Classifier(PCPCLSFR_NAME)
    ct_clsfr = Classifier(CTCLSFR_NAME)
    clsfr_mgr = ClassificationMgr([pub_clsfr, ct_clsfr])
    return clsfr_mgr


def create_classifier():
    """Create classifier obj."""
    ct_clsfr = Classifier(CTCLSFR_NAME)
    ct_clsfr.load()
    return ct_clsfr


def process_records(offset, chunksize, daomgr: DaoManager, classifier: Classifier):
    """Download predict and save records."""
    try:
        pid = os.getpid()
        print(f"Process: {pid} - Downloading {chunksize} features")
        features_df = daomgr.get_features(offset, chunksize)
        if not features_df.empty:
            print(f"Process: {pid} - Calculating {chunksize} predictions")
            predicted_df = classifier.predict(features_df)
            print(f"Process: {pid} - Saving {chunksize} predictions")
            daomgr.save_predictions(predicted_df)
        return True
    except Exception as exp:  # noqa
        sys.stderr.write(
            f"Unable to process records: {offset} to  {offset + chunksize}"
        )
        sys.stderr.write(str(exp))


def run_predictions():
    """Calculate predictions for all records in database."""
    warnings.filterwarnings("ignore")
    daomgr: DaoManager = create_daomgr()
    classifier = create_classifier()
    chunksize = config.get_chunksize()
    start_time = time()
    max_val = daomgr.get_max_feature(chunksize)
    offsets = range(0, max_val, chunksize)
    max_workers = config.get_max_workers()
    outputs = list()
    with ProcessPoolExecutor(max_workers=max_workers) as executor:
        for out in executor.map(
            process_records,
            offsets,
            repeat(chunksize),
            repeat(daomgr),
            repeat(classifier),
        ):
            outputs.append(out)
    elapsed_time = round(time() - start_time)
    print(f"Elapsed time: {elapsed_time}s")


def create_tables():
    """Call scrapers to create all tables."""
    daomgr: DaoManager = create_daomgr()
    chunksize = config.get_chunksize()
    daomgr.download(chunksize, baseline=True)


def update_tables():
    """Call scrapes to download data and create/append tables."""
    daomgr: DaoManager = create_daomgr()
    chunksize = config.get_chunksize()
    daomgr.download(chunksize)


def sample_run():
    """Run through entire cycle of creating tables and predictions with sample data."""  # noqa
    daomgr: DaoManager = create_daomgr()
    chunksize = config.get_chunksize()
    daomgr.download(chunksize, is_sample=True, baseline=True)


def init():
    """Create configuration file if doesn't exist."""
    assert config.CONFIG_FILE.exists()
    print(f"GeniePy version {__version__} is ready")



"""PubMed related functionality."""
import xml.etree.ElementTree as ET
import json
import csv
import jsonlines


class PubMedArticle:
    """
    PubMed Article model.

    Reference documentation:
    https: // www.nlm.nih.gov/bsd/licensee/elements_descriptions.html
    """

    MEDLINE_TAG = "MedlineCitation"
    ARTICLE_TAG = MEDLINE_TAG + "/Article"
    JOURNAL_TAG = ARTICLE_TAG + "/Journal"

    __slots__ = ["_pubmed_article", "citationCount", "citationPmid"]

    def __init__(self, article_tree: ET.Element):
        """Construct object from corresponding article element tree."""
        self._pubmed_article: ET.Element = article_tree
        self.citationCount: int = 0
        self.citationPmid: str = ""

    def _get_xml_element(self, tags: [str], tag_attrib=None) -> str:
        """
        Retrieve xml element given full element xml path.

        Arguments:
            element [str] -- list of xml sub tags to target elmeent
                                i.e. ['MedlineCitation', 'PMID']
            tag_attib -- Optional attribute if retrieving atrrib

        Returns:
            str -- The text value of the queried element. If element doesn't
                    exist returns empty string.
        """
        element_path = ""
        for tag in tags:
            element_path += tag + "/"
        element_path = element_path[:-1]  # remove last '/'
        element = self._pubmed_article.find(element_path)
        if element is None:
            return ""

        if tag_attrib is None:
            return element.text
        else:
            return element.attrib[tag_attrib]

    @property
    def pmid(self) -> str:
        """Pubmed article ID."""
        pmid = self._get_xml_element([self.MEDLINE_TAG, "PMID"])
        return pmid

    @property
    def date_completed(self) -> str:
        """Date completed record distributed to PubMed."""
        year = self._get_xml_element([self.MEDLINE_TAG, "DateCompleted", "Year"])
        month = self._get_xml_element([self.MEDLINE_TAG, "DateCompleted", "Month"])
        day = self._get_xml_element([self.MEDLINE_TAG, "DateCompleted", "Day"])
        return year + "-" + month + "-" + day

    @property
    def pub_model(self) -> str:
        """Publication model - medium/media in which article was published."""
        pub_model = self._get_xml_element([self.ARTICLE_TAG], tag_attrib="PubModel")
        return pub_model

    @property
    def title(self) -> str:
        """Full journal title."""
        title = self._get_xml_element([self.JOURNAL_TAG, "Title"])
        return title

    @property
    def iso_abbreviation(self) -> str:
        """Journal title ISO abbreviation."""
        iso_abbrev = self._get_xml_element([self.JOURNAL_TAG, "ISOAbbreviation"])
        return iso_abbrev

    @property
    def article_title(self) -> str:
        """Entire title of journal article in English."""
        article_title = self._get_xml_element([self.ARTICLE_TAG, "ArticleTitle"])
        return article_title

    @property
    def abstract(self) -> str:
        """Entire abstract taken directly from published article."""
        abstract = self._get_xml_element([self.ARTICLE_TAG, "Abstract", "AbstractText"])
        return abstract

    @property
    def authors(self) -> [str]:
        """Names of authors published with article."""
        authors: [str] = []
        author_list = self._pubmed_article.findall(
            self.ARTICLE_TAG + "/AuthorList/Author"
        )
        for author in author_list:
            lastname_el = author.find("LastName")
            author_lastname = "" if lastname_el is None else lastname_el.text

            forename_el = author.find("ForeName")
            author_forename = "" if forename_el is None else forename_el.text

            authors.append(author_lastname + ", " + author_forename)
        return authors

    @property
    def language(self) -> str:
        """Tha language the article was published in."""
        language = self._get_xml_element([self.ARTICLE_TAG, "Language"])
        return language

    @property
    def chemicals(self) -> [str]:
        """One or more chemical elements."""
        chemicals: [str] = []
        chemicals_list = self._pubmed_article.findall(
            self.MEDLINE_TAG + "/ChemicalList/Chemical"
        )
        for chemical in chemicals_list:
            chemical_substance = chemical.find("NameOfSubstance")
            if chemical_substance is not None:
                chemicals.append(chemical_substance.text)
        return chemicals

    @property
    def mesh_list(self) -> [str]:
        """Article's suppl mesh list."""
        meshes: [str] = []
        mesh_list = self._pubmed_article.findall(
            self.MEDLINE_TAG + "/MeshHeadingList/MeshHeading"
        )
        for mesh in mesh_list:
            descriptor = mesh.find("DescriptorName")
            if descriptor is not None:
                meshes.append(descriptor.text)
        return meshes

    @property
    def issn(self) -> str:
        """Journal ISSN ID."""
        issn = self._get_xml_element([self.JOURNAL_TAG, "ISSN"])
        return issn

    @property
    def issn_type(self) -> str:
        """Journal ISSN Type."""
        issn_type = self._get_xml_element(
            [self.JOURNAL_TAG, "ISSN"], tag_attrib="IssnType"
        )
        return issn_type

    @property
    def citation_count(self) -> str:
        """Journal ISSN Type."""
        issn_type = self._get_xml_element(
            [self.JOURNAL_TAG, "ISSN"], tag_attrib="IssnType"
        )
        return issn_type

    def set_citationCount(self, citation_count: int):
        """Update property: citation_count."""
        self.citationCount = citation_count

    def set_citationPmid(self, citation_pmid: str):
        """Update property: citation_pmid."""
        self.citationPmid = citation_pmid

    @property
    def to_dict(self):
        """Generate article model dictionary."""
        _dict = {}
        _dict["pmid"] = self.pmid
        _dict["date_completed"] = self.date_completed
        _dict["pub_model"] = self.pub_model
        _dict["title"] = self.title
        _dict["iso_abbreviation"] = self.iso_abbreviation
        _dict["article_title"] = self.article_title
        _dict["abstract"] = self.abstract
        _dict["authors"] = self.authors
        _dict["language"] = self.language
        _dict["chemicals"] = self.chemicals
        _dict["mesh_list"] = self.mesh_list
        _dict["issn"] = self.issn
        _dict["issn_type"] = self.issn_type
        _dict["citation_count"] = self.citationCount
        _dict["citation_pmid"] = self.citationPmid
        return _dict


class ArticleSetParser:
    """
    Parsing utility for PubMed Article sets.

    An article set is an xml file with an array of <PubMedArticles>.
    """

    @staticmethod
    def extract_articles(xml_file_path: str) -> [PubMedArticle]:
        """
        Extract articles from xml file.

        Arguments:
            xml_file_path {str} -- absolute path to xml file

        Returns:
            [PubMedArticle] -- List of pubmed article objects
        """
        xml_root: ET.Element = ET.parse(xml_file_path).getroot()
        articles_xml_list = xml_root.findall("PubmedArticle")
        pubmed_articles: [PubMedArticle] = []
        for article_xml in articles_xml_list:
            pubmed_articles.append(PubMedArticle(article_xml))
        return pubmed_articles

    @staticmethod
    def articles_to_dict(articles: [PubMedArticle]) -> [dict]:
        """Generate list of dictionaries from articles."""
        dict_list = []
        for article in articles:
            dict_list.append(article.to_dict)
        return dict_list

    @staticmethod
    def articles_to_json(articles: [PubMedArticle], target_file_path: str):
        """Serialize pubmedarticle objects to json file."""
        dict_list = ArticleSetParser.articles_to_dict(articles)
        articles_json = json.dumps({"articles": dict_list})
        with open(target_file_path, "w") as target_file:
            target_file.write(articles_json)

    @staticmethod
    def articles_to_jsonl(articles: [PubMedArticle], target_file_path: str):
        """Serialize pubmedarticle objects to jsonl file."""
        dict_list = ArticleSetParser.articles_to_dict(articles)
        with jsonlines.open(target_file_path, "w") as writer:
            # pylint: disable=E1101
            writer.write_all(dict_list)

    @staticmethod
    def articles_to_pipe(articles: [PubMedArticle], target_file_path: str):
        """Serialize pubmedarticle objects to csv file."""
        dict_list = ArticleSetParser.articles_to_dict(articles)
        csv_columns = [
            "pmid",
            "date_completed",
            "pub_model",
            "title",
            "iso_abbreviation",
            "article_title",
            "abstract",
            "authors",
            "language",
            "chemicals",
            "mesh_list",
            "issn",
            "issn_type",
            "citation_count",
            "citation_pmid",
        ]

        try:
            with open(target_file_path, "w", encoding="utf-8") as csv_file:
                writer = csv.DictWriter(csv_file, delimiter="|", fieldnames=csv_columns)
                writer.writeheader()
                for article in dict_list:
                    writer.writerow(article)
        except IOError:  # pragma: no cover
            print("Unable to write csv file")



"""Scraping module to fetch data from online sources."""
import binascii
from pathlib import Path
from typing import Generator
from abc import ABC, abstractmethod
import wget
import gzip
import shutil
import os
import pandas as pd
import geniepy.config as config
from ftplib import FTP
import requests
import xml.etree.ElementTree as ET
from random import randint
from pathlib import Path
from geniepy.pubmed import PubMedArticle
from joblib import Memory


# Initialize memoization
memory = Memory(str(config.TMP_DIR), verbose=0)


class BaseScraper(ABC):
    """Scraper Abstract Base Class."""

    @abstractmethod
    def scrape(self, chunksize: int, **kwargs) -> Generator:
        """
        Download any new information not yet downloaded.

        The first time this method is called, it should download and return
        all available information from the online source to build the internal
        databases. Subsequent calls to the scrape function should only return
        data that is new and has not yet been downloaded.

        Since the data downloaded can be too large to be loaded into memory at
        once, it should be yielded incrementally in chunks. Refer to the chunksize
        argument to determine how much data should be yielded at a time.

        Keyword Arguments:
            chunksize {int} -- the size of each chunk the data should be returned

        Returns:
            Generator -- The generator yielding the data in given chunksizes.
        """


class PubtatorGeneScraper(BaseScraper):
    """Scrape PMID/GENEID from pubtator."""

    SOURCE_NAME = "Pubtator Genes"
    TMP_DIR = config.TMP_DIR
    FTP_URL = "ftp://ftp.ncbi.nlm.nih.gov/pub/lu/PubTatorCentral/gene2pubtatorcentral.gz"  # noqa
    SAMPLE_FTP_URL = "ftp://ftp.ncbi.nlm.nih.gov/pub/lu/PubTatorCentral/gene2pubtatorcentral.sample"  # noqa
    DOWNLOAD_NAME = "pubtator-gene.download"
    CSV_NAME = "pubtator-gene.csv"
    HEADER_NAMES = ["PMID", "Type", "GeneID", "Mentions", "Resource"]

    def __init__(self):
        self.TMP_DIR.mkdir(parents=True, exist_ok=True)

    @property
    def download_path(self):
        """Path to pubtator gzip downloaded file."""
        download_path = self.TMP_DIR.joinpath(self.DOWNLOAD_NAME).resolve()
        return download_path

    @property
    def csv_path(self):
        """Path to pubtator csv downloaded file."""
        csv_path = self.TMP_DIR.joinpath(self.CSV_NAME).resolve()
        return csv_path

    def is_gzip(self, filepath) -> bool:
        """Check if file is gzipped."""
        with open(filepath, "rb") as test_f:
            return binascii.hexlify(test_f.read(2)) == b"1f8b"

    def download(self):
        """Download records from online sources."""
        print(f"Downloading {self.SOURCE_NAME} records...")
        if not self.download_path.exists():
            wget.download(self.FTP_URL, str(self.download_path))
        if self.is_gzip(self.download_path):
            with gzip.open(self.download_path, "rb") as f_in:
                with open(self.csv_path, "wb") as f_out:
                    shutil.copyfileobj(f_in, f_out)
        else:
            self.download_path.rename(self.csv_path)
        self.clean_up()

    def set_sample(self, kwargs):
        """Set sample url if is_sample is set."""
        if kwargs.get("is_sample"):
            self.FTP_URL = self.SAMPLE_FTP_URL

    def scrape(self, chunksize: int, **kwargs) -> Generator:
        """
        Download pmid/geneid data from pubtator.

        ftp://ftp.ncbi.nlm.nih.gov/pub/lu/PubTatorCentral/disease2pubtatorcentral.gz
        """
        self.set_sample(kwargs)
        if kwargs.get("baseline"):
            self.download()
            csv_gen = pd.read_csv(
                self.csv_path,
                chunksize=chunksize,
                delimiter="\t",
                names=self.HEADER_NAMES,
            )
            return csv_gen
        return ()

    def clean_up(self):
        """Delete all temp files."""
        if os.path.exists(self.download_path):
            os.remove(self.download_path)


class PubtatorDiseaseScraper(PubtatorGeneScraper):
    """Scrape PMID/DiseaseID from pubtator."""

    SOURCE_NAME = "Pubtator Diseases"
    FTP_URL = "ftp://ftp.ncbi.nlm.nih.gov/pub/lu/PubTatorCentral/disease2pubtatorcentral.gz"  # noqa
    SAMPLE_FTP_URL = "ftp://ftp.ncbi.nlm.nih.gov/pub/lu/PubTatorCentral/disease2pubtatorcentral.sample"  # noqa
    DOWNLOAD_NAME = "pubtator-disease.download"
    CSV_NAME = "pubtator-disease.csv"
    HEADER_NAMES = ["PMID", "Type", "DiseaseID", "Mentions", "Resource"]


class SjrScraper(PubtatorGeneScraper):
    """Scrape Scientific Journal Ratings records."""

    SOURCE_NAME = "SJR"
    FTP_URL = "https://www.scimagojr.com/journalrank.php?out=xls"  # noqa
    SAMPLE_FTP_URL = "https://www.scimagojr.com/journalrank.php?area=1100&out=xls"
    CSV_NAME = "sjr.csv"

    def scrape(self, chunksize: int, **kwargs) -> Generator:
        """Download sjr data."""
        self.set_sample(kwargs)
        if kwargs.get("baseline") is True:
            # Only download if baseline is true
            self.download()
            csv_gen = pd.read_csv(self.csv_path, chunksize=chunksize, delimiter=";")
            return csv_gen
        return ()


class CtdScraper(BaseScraper):
    """
    Implementation of CTD Gene-Disease Relationship Scraper.

    http://ctdbase.org/
    """

    def scrape(self, chunksize: int, **kwargs) -> Generator:
        """
        Implement base scrape method to download records from CTD database.

        Keyword Arguments:
            chunksize {int} -- the size of each chunk the data should be returned

        Returns:
            Generator -- The generator yielding the data in given chunksizes.
        """
        raise NotImplementedError


class PubMedScraper(BaseScraper):
    """
    Implementation of PubMed Articles Scraper.

    https://www.ncbi.nlm.nih.gov/pubmed/
    https://www.nlm.nih.gov/bsd/licensee/elements_descriptions.html

    """

    # Logger
    LOGGER = config.get_logger("PubMedScraper")

    # Constants for PubMed scraping
    DEFAULT_CHUNKSIZE: int = 1000
    DEFAULT_PUBMED_BASELINE_SCRAPE_MODE: bool = False
    DEFAULT_PUBMED_FTP_SERVER: str = "ftp.ncbi.nlm.nih.gov"
    DEFAULT_PUBMED_BASELINE_DIR: str = "/pubmed/baseline/"
    DEFAULT_PUBMED_UPDATE_DIR: str = "/pubmed/updatefiles/"
    DEFAULT_PUBMED_DATA_FILE = "~/.geniepy.d/pubmed.dat"
    DEFAULT_DOWNLOAD_DIR = "~/.geniepy.d/tmp/"
    DEFAULT_PUBMED_DATAFILE_EXTN = ".xml.gz"
    DEFAULT_DOWNLOAD_RETRIES = 2

    # Constants for PubMed scraping
    TAG_ARTICLE = "PubmedArticle"

    # Constants for Citation scraping
    API_BASE_URL = "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/elink.fcgi?dbfrom=pubmed&linkname=pubmed_pubmed_citedin"  # noqa
    API_ID_PARAM = "&id="
    API_KEY_PARAM = "&api_key="
    API_KEYS = [
        "70b53d4e84436970587aef3493a723cae708",
        "9ab4b04dcabcab740d1a297e6f3f54aa1e09",
        "39458b83e654b32523475ee6b828b6c22c08",
        "cf95d471b41c4ea9e6b89d41de269705eb08",
        "94c3049a7b3dffca2d22ce11b51a3ea5da09",
        "3fb37654e8f711f8751d9e3c21fc46257108",
        "91315e4e69c1b5e2676371490cfc715caa08",
        "b1ee76fae01c312e4bf2fd3e8b1bcf978a09",
        "25b74bb456d7e6fc756c86c7aebb1b525009",
        "7a7e4492084dd09cef758ee65a7909704b08",
        "4d32405f13fed69766ad639d89c5c5288608",
        "18c8d26e60b5cd012e2ba9618c88dd12cb09",
        "667ca5b1b72b7697ed43c26727329282bb08",
        "5cdea4b682e2fc93c56c2b1e5e9da93f4709",
        "108b2e1a81369991da6c737545097c581a08",
        "f10805158a7609bb115cb074b05e5923a407",
        "00d74dd1cb732cd7fc29f54168fe7055c809",
    ]
    TAG_CITATION_ID = "./LinkSet/LinkSetDb/Link/Id"

    def __init__(self):
        """Initialize memoization for citation scraping."""
        self._citationScrape = memory.cache(self._citationScrape)

    def scrape(self, chunksize: int, **kwargs) -> Generator:
        """
        Implement base scrape method to download records from PubMed database.

        Keyword Arguments:
            chunksize {int} -- the size of each chunk the data should be returned

        Returns:
            Generator -- The generator yielding the data in given chunksizes.
        """
        PubMedScraper.LOGGER.info("Starting PubMed Scraper")

        # set default chunksize
        if chunksize <= 0:
            chunksize = PubMedScraper.DEFAULT_CHUNKSIZE
            PubMedScraper.LOGGER.info(f"chunksize set to default: {chunksize}")

        # set IS_SAMPLE flag
        IS_SAMPLE = False
        if "is_sample" in kwargs:
            try:
                IS_SAMPLE = kwargs.get("is_sample")
            except:
                IS_SAMPLE = False

        # set BASELINE_SCRAPE_MODE flag
        # BASELINE_SCRAPE_MODE = True (enables the baseline scraping mode)
        # BASELINE_SCRAPE_MODE = Flase (enables the update scraping mode)
        BASELINE_SCRAPE_MODE: bool = PubMedScraper.DEFAULT_PUBMED_BASELINE_SCRAPE_MODE
        if "baseline" in kwargs:
            try:
                if kwargs.get("baseline") == True:
                    BASELINE_SCRAPE_MODE = True
                    PubMedScraper.LOGGER.info("Scrape mode: Baseline")
            except Exception as e:
                BASELINE_SCRAPE_MODE = PubMedScraper.DEFAULT_PUBMED_BASELINE_SCRAPE_MODE
                PubMedScraper.LOGGER.exception(e.msg())
                PubMedScraper.LOGGER.info("Scrape mode: Daily Update")
        else:
            PubMedScraper.LOGGER.info("Scrape mode: Daily Update")

        # set PubMed FTP server
        FTP_SERVER: str = ""
        try:
            FTP_SERVER = config.get_pubmed_ftp_server()
        except Exception as e:
            FTP_SERVER = PubMedScraper.DEFAULT_PUBMED_FTP_SERVER
            PubMedScraper.LOGGER.exception(e.msg())

        PubMedScraper.LOGGER.info(f"FTP_SERVER: {FTP_SERVER}")

        # set PubMed FTP directory
        FTP_DIR: str = ""
        if BASELINE_SCRAPE_MODE:
            # we are in baseline scrape mode. get baseline dataset directory
            try:
                FTP_DIR = config.get_pubmed_baseline_dir()
            except Exception as e:
                FTP_DIR = PubMedScraper.DEFAULT_PUBMED_BASELINE_DIR
                PubMedScraper.LOGGER.exception(e.msg())
        else:
            # we are in update scrape mode. get baseline dataset directory
            try:
                FTP_DIR = config.get_pubmed_update_dir()
            except Exception as e:
                FTP_DIR = PubMedScraper.DEFAULT_PUBMED_UPDATE_DIR
                PubMedScraper.LOGGER.exception(e.msg())

        PubMedScraper.LOGGER.info(f"FTP_DIR: {FTP_DIR}")

        # connect to pubmed ftp and retrieve list of ftp files
        pubmed_ftp = self._ftp_connect(FTP_SERVER, FTP_DIR)
        pubmed_files = self._ftp_file_list(pubmed_ftp)  # list of files from ftp
        pubmed_history = self._read_history()  # list of historically scraped files

        PubMedScraper.LOGGER.info(f"Number of files in FTP: {len(pubmed_files)}")
        PubMedScraper.LOGGER.info(f"Number of files in History: {len(pubmed_history)}")

        # determine new files to parse
        # in BASELINE_SCRAPE_MODE parse all files
        pubmed_new_files = []
        parsed_files = []
        if BASELINE_SCRAPE_MODE:
            pubmed_new_files = pubmed_files
        else:
            pubmed_new_files = list(set(pubmed_files) - set(pubmed_history))

        PubMedScraper.LOGGER.info(
            f"Number of new files to be parsed: {len(pubmed_new_files)}"
        )

        # main scraping block
        try:
            for pubmed_file in pubmed_new_files:
                # download pubmed data file
                retries = 0
                while retries <= PubMedScraper.DEFAULT_DOWNLOAD_RETRIES:
                    if self._ftp_download(pubmed_ftp, pubmed_file):
                        PubMedScraper.LOGGER.info(
                            f"Downloaded PubMed file: {pubmed_file}"
                        )
                        break
                    retries += 1

                # scrape downloaded PubMed data file
                articles = self._pubmedScrape(pubmed_file)
                PubMedScraper.LOGGER.info(
                    f"{len(articles)} articles found in {pubmed_file}"
                )

                # optimize chunksize based on number of
                # available PubMed articles
                chunk_size: int = 0
                if chunksize > len(articles):
                    chunk_size = len(articles)
                else:
                    chunk_size = chunksize
                PubMedScraper.LOGGER.info(f"Chunksize is set to: {chunk_size}")

                # yield articles to generator
                while True:
                    articles_chunk = []
                    for i in range(chunk_size):
                        if len(articles) > 0:
                            article = articles.pop()
                            if not IS_SAMPLE or (IS_SAMPLE and i <= 1000):
                                # scrape citation metadata
                                # in is_sample mode only scrape first 1000 citations
                                citations = self._citationScrape(article.pmid)
                                article.set_citationCount(citations[1])
                                article.set_citationPmid(citations[2])
                                PubMedScraper.LOGGER.info(
                                    f"Get citations for PMID: {article.pmid} [{i} of {chunk_size}]"
                                )
                            articles_chunk.append(article)
                        else:
                            break

                    if len(articles_chunk) <= 0:
                        break

                    PubMedScraper.LOGGER.info(f"Yielded {len(articles_chunk)} articles")
                    yield articles_chunk

                # update list of parsed files
                parsed_files.append(pubmed_file)

        except GeneratorExit:
            # ignore error.
            pass

        finally:
            # close ftp connection
            self._ftp_disconnect(pubmed_ftp)
            PubMedScraper.LOGGER.info("FTP disconnected")

            # clean up downloaded files
            self._clean_up()
            PubMedScraper.LOGGER.info("PubMed Scraper: Clean-up")

            # update history
            self._update_history(parsed_files)
            PubMedScraper.LOGGER.info("PubMed Scraper: Updated History")

            # if baseline dataset was scraped; reset scrape history
            if BASELINE_SCRAPE_MODE:
                self._clear_history()
                PubMedScraper.LOGGER.info(
                    "PubMed Scraper: Cleared History (after BASELINE)"
                )

        return

    def _ftp_connect(self, ftp_server, ftp_dir) -> FTP:
        """Opens FTP connection and returns the FTP handle"""
        try:
            ftp = FTP(ftp_server)
            ftp.login()
            ftp.cwd(ftp_dir)
            return ftp
        except Exception as e:
            PubMedScraper.LOGGER.exception(e.msg())
            return None

    def _ftp_disconnect(self, ftp: FTP):
        """Disconnect and close FTP connection"""
        try:
            ftp.close()
            return
        except Exception as e:
            PubMedScraper.LOGGER.exception(e.msg())
            return

    def _ftp_file_list(self, ftp: FTP) -> []:
        """Read list of files in FTP directory"""
        ftp_files = []
        try:
            ftp.retrlines(
                f"LIST *{PubMedScraper.DEFAULT_PUBMED_DATAFILE_EXTN}", ftp_files.append
            )
            ftp_files = [
                file.split(" ")[-1].lower() for file in ftp_files
            ]  # transform filenames
            return ftp_files
        except Exception as e:
            PubMedScraper.LOGGER.exception(e.msg())
            return ftp_files

    def _ftp_download(self, ftp: FTP, ftp_file: str) -> bool:
        """Downloads given file from the FTP server"""
        try:
            # check if the download path exists.
            # if not, create the path
            download_path = os.path.expanduser(self._get_download_dir())
            Path(download_path).mkdir(parents=True, exist_ok=True)

            # downlad file
            download_filepath = os.path.join(download_path, ftp_file)
            ftp.retrbinary("RETR " + ftp_file, open(download_filepath, "wb").write)
            return True
        except Exception as e:
            PubMedScraper.LOGGER.exception(e.msg())
            return False

    def _read_history(self) -> []:
        """Read and create list of PubMed filenames
        which are already scraped"""
        history = []
        try:
            with open(self._get_history_filepath(), "r") as f:
                for line in f.readlines():
                    history.append(line.replace("\n", "").lower())
            return history
        except Exception as e:
            PubMedScraper.LOGGER.info(
                "History file not found. New History file will be created."
            )
            return []

    def _update_history(self, file_list: []):
        """Read and create list of PubMed filenames
        which are already scraped"""
        try:
            with open(self._get_history_filepath(), "a+") as f:
                f.writelines(map(lambda x: x + "\n", file_list))
        except Exception as e:
            PubMedScraper.LOGGER.exception(e.msg())
            return

    def _clear_history(self):
        """Clear scrape history"""
        history_file = self._get_history_filepath()
        if os.path.exists(history_file):
            os.remove(history_file)
        pass

    def _get_history_filepath(self) -> str:
        """Get PubMed data file path from Config"""
        try:
            history_file_path = os.path.expanduser(config.get_pubmed_data_file())
            return history_file_path
        except Exception as e:
            PubMedScraper.LOGGER.exception(e.msg())
            return PubMedScraper.DEFAULT_PUBMED_BASELINE_DIR

    def _get_download_dir(self) -> str:
        """Get path where to download PubMed data files"""
        try:
            return config.get_pubmed_download_dir()
        except Exception as e:
            PubMedScraper.LOGGER.exception(e.msg())
            return PubMedScraper.DEFAULT_DOWNLOAD_DIR

    def _pubmedScrape(self, pubmed_file) -> []:
        """Scrape all pubmed articles from pubmed data file"""
        pubmed_articles = []
        try:
            file_path = os.path.expanduser(self._get_download_dir())
            file = os.path.join(file_path, pubmed_file)
            with gzip.open(file, "rb") as f:
                xml_root = ET.fromstring(f.read())
                xml_list = xml_root.findall(PubMedScraper.TAG_ARTICLE)
                for article_xml in xml_list:
                    pubmed_articles.append(PubMedArticle(article_xml))
            return pubmed_articles
        except Exception as e:
            PubMedScraper.LOGGER.exception(e.msg())
            return []

    def _citationScrape(self, pmid: int) -> []:
        """Scrape citation metadata from PubMed API"""
        null_citation = [pmid, 0, ""]

        # fire PubMed API request
        try:
            req = requests.get(PubMedScraper._citationApiUrl(pmid))
            tree = ET.fromstring(req.text)
        except Exception as e:
            PubMedScraper.LOGGER.exception(e.msg())
            return null_citation

        # scrape citation data from API response
        citation_tree = tree.findall(PubMedScraper.TAG_CITATION_ID)
        if citation_tree and len(citation_tree) > 0:
            citations = []
            for _ in citation_tree:
                citations.append(_.text)
            return [pmid, len(citation_tree), ",".join(citations)]

        return null_citation

    def _clean_up(self):
        """Delete all downloaded pubmed files"""
        clean_up_path = os.path.expanduser(self._get_download_dir())
        try:
            shutil.rmtree(clean_up_path)
            return
        except OSError as e:
            PubMedScraper.LOGGER.exception(e.msg())
            return

    @classmethod
    def _citationApiUrl(cls, pmid: int) -> str:
        """Returns a fully formated citation API URL for a given PMID"""
        return (
            PubMedScraper.API_BASE_URL
            + PubMedScraper.API_KEY_PARAM
            + PubMedScraper._randomApiKey()
            + PubMedScraper.API_ID_PARAM
            + str(pmid)
        )

    @classmethod
    def _randomApiKey(cls) -> str:
        """Returns a random PubMed API key from a list of static API keys"""
        return PubMedScraper.API_KEYS[randint(0, len(PubMedScraper.API_KEYS) - 1)]



"""Test base classifier class."""
import pandas as pd
import geniepy.classmgmt.classifiers as clsfr

CLSFR_NAME = "pub_score"

SAMPLE_RECORD = pd.Series(["g", "e", "e", "k", "s"])


class TestClassifier:
    """Base Classifier pytest class."""

    classifier = clsfr.Classifier("pub_score")

    def test_constructor(self):
        """Test classifier object is created."""
        assert self.classifier is not None
        assert self.classifier.name == CLSFR_NAME
        assert self.classifier.is_trained is False

    def test_predict_not_trained(self):
        """Brand new classifier is not trained, should return -1."""
        self.classifier = clsfr.Classifier(CLSFR_NAME)
        assert self.classifier.is_trained is False
        prediction = self.classifier.predict(None)
        assert isinstance(prediction, float)
        assert prediction == clsfr.ERROR_SCORE

    def test_predict_none(self):
        """Predict should always return a number so doesn't halt classmgr."""
        prediction = self.classifier.predict(None)
        assert isinstance(prediction, float)
        assert prediction == clsfr.ERROR_SCORE



"""Integration tests of Pubtator - Gene data."""
import pytest
from pathlib import Path
from geniepy.datamgmt.scrapers import PubtatorGeneScraper
from tests.resources.mock import TEST_CHUNKSIZE


FTP_URL_SAMPLE = (
    "ftp://ftp.ncbi.nlm.nih.gov/pub/lu/PubTatorCentral/gene2pubtatorcentral.gz"
)


@pytest.mark.slow_integration_test
class TestPubtatorGeneScraper:
    """Test pubtator gene scraper."""

    def test_download():
        """Test downloding data."""
        pbs = PubtatorGeneScraper()
        pbs.clean_up()
        pbs.download()
        assert Path.exists(pbs.PUBTATOR_CSV_NAME)
        assert Path.exists(pbs.PUBTATOR_GZIP_NAME)

    def test_cleanup():
        """Test deleting generated files."""
        pbs = PubtatorGeneScraper()
        if not Path.exists(pbs.PUBTATOR_CSV_NAME):
            Path(pbs.PUBTATOR_CSV_NAME).touch()
        pbs.clean_up()
        assert not Path.exists(pbs.PUBTATOR_CSV_NAME)

    def test_scrape():
        """Test scraping data from pubtator-gene."""
        pbs = PubtatorGeneScraper()
        gen = pbs.scrape(TEST_CHUNKSIZE)
        df = next(gen)
        assert df.shape[0] == TEST_CHUNKSIZE



"""Basic pytest test to ensure testing framework is working."""
import pytest


def test_pytest():
    """Test is pytest is working."""
    with pytest.raises(Exception):
        raise Exception
    assert True



"""
Module to bootstrap genie to scripts.

Scripts in this directory may call genie packages, this module adds genie
top-level dir to python path for scripts to be able to import genie.
"""
import sys
from pathlib import Path

GENIE_PATH = Path(__file__).resolve().parents[2]
sys.path.append(str(GENIE_PATH))



"""Data Access Repositories to abstract interation with databases."""
from typing import Generator
from abc import ABC, abstractmethod
import pandas as pd
from datetime import datetime
from pandas import DataFrame
from google.oauth2 import service_account
import pandas_gbq
from sqlalchemy import create_engine, Table
from geniepy.errors import DaoError, ConnectionError
from geniepy.datamgmt.tables import RepoProperties


class BaseRepository(ABC):
    """Base Abstract Class for Data Access Object Repositories."""

    __slots__ = ["_table", "_tablename", "_pkey"]

    @property
    def tablename(self) -> str:
        """Return DAO repo's tablename."""
        # pylint: disable=no-member
        return self._tablename

    @property
    def query_all(self) -> str:
        """Generate query string to query entire table."""
        return f"SELECT * FROM {self.tablename}"

    def query_pkey(self, val) -> str:
        """Generate query string to query by primary key."""
        if isinstance(val, str):
            # pylint: disable=no-member
            return f"SELECT * FROM {self.tablename} WHERE {self._pkey}='{val}'"
        else:
            # Don't need quotes surrounding val if not a str
            # pylint: disable=no-member
            return f"SELECT * FROM {self.tablename} WHERE {self._pkey}={val}"

    @abstractmethod
    def save(self, payload: DataFrame):
        """
        Save payload to database table.

        Arguments:
            payload {DataFrame} -- the payload to be stored in db

        Raises:
            DaoError: if cannot save payload to db
        """

    @abstractmethod
    def delete_all(self):
        """Delete all records in repository."""

    @abstractmethod
    # pylint: disable=bad-continuation
    def query(
        self, query: str, chunksize: int, exact=False
    ) -> Generator[DataFrame, None, None]:
        """
        Query DAO repo and returns a generator of DataFrames with query results.

        Keyword Arguments:
            query {str} -- Query string
            chunksize {int} -- Number of rows of dataframe per chunk
            exact {bool} -- If exact is false, query function manages chunks
                and returns generator. Otherwise, the direct query is returned.

        Returns:
            Generator[DataFrame] -- Generator to iterate over DataFrame results,
                or DataFrame if exact param is True
        """


class SqlRepository(BaseRepository):
    """Implementation of Sqlite Data Access Object Repository."""

    __slots__ = ["_engine"]

    def __init__(self, db_loc: str, propty: RepoProperties):
        """
        Initialize DAO repository and create table.

        Arguments:
            db_loc {str} -- location of underlying database
            propty {RepoProperties} -- Repository properties structure
        """
        self._tablename = propty.tablename
        self._table = propty.table
        self._pkey = propty.pkey
        # Create sql engine
        self._engine = create_engine(db_loc)
        # Create Table
        self._table.create(self._engine)

    def save(self, payload: DataFrame):
        """
        Save payload to database table.

        Arguments:
            payload {DataFrame} -- the payload to be stored in db

        Raises:
            DaoError: if cannot save payload to db
        """
        try:
            payload.to_sql(
                self._tablename,
                con=self._engine,
                if_exists="append",
                index=False,
                method="multi",
            )
        except Exception as sql_exp:
            raise DaoError(sql_exp)

    def delete_all(self):
        """Delete all records in repository."""
        self._table.drop(self._engine)
        self._table.create(self._engine)

    # pylint: disable=bad-continuation
    def query(
        self, query: str, chunksize: int, exact=False
    ) -> Generator[DataFrame, None, None]:
        """
        Query DAO repo and returns a generator of DataFrames with query results.

        Keyword Arguments:
            query {str} -- Query string
            chunksize {int} -- Number of rows of dataframe per chunk
            exact {bool} -- If exact is false, query function manages chunks
                and returns generator. Otherwise, the direct query is returned.

        Returny:
            Generator[DataFrame] -- Generator to iterate over DataFrame results.
        """
        if query is None:
            raise DaoError
        # If query string provided
        generator = pd.read_sql_query(query, self._engine, chunksize=chunksize)
        return generator


class GbqRepository(BaseRepository):  # pragma: no cover
    """Implementation of Sqlite Data Access Object Repository."""

    __slots__ = ["_proj", "_credentials_path"]

    # pylint: disable=bad-continuation
    def __init__(
        self, db_loc: str, propty: RepoProperties, dataset: str, credentials: str
    ):
        """
        Initialize DAO repository and create table.

        Arguments:
            db_loc {str} -- name of BigQuery project
            propty {RepoProperties} -- Repository properties structure
            dataset {str} -- The GBQ dataset the table belongs to
            credentials_path {str} -- path to gcp credentials json
        """
        self._proj = db_loc
        self._tablename = dataset + "." + propty.tablename
        self._pkey = propty.pkey
        self._table = self.get_dict_schema(propty.table)
        self._credentials_path = credentials
        self.connect()

    def connect(self):
        """Connect to Google BigQuery."""
        try:
            bgq_credentials = service_account.Credentials.from_service_account_file(
                self._credentials_path,
            )
            pandas_gbq.context.credentials = bgq_credentials
            pandas_gbq.context.project = self._proj
        except Exception as exp:
            raise ConnectionError(exp)

    @staticmethod
    def get_dict_schema(table_schema: Table) -> [dict]:
        """Convert Table Schema into GBQ dict expected format."""
        # flake8: noqa
        if table_schema is None:
            return
        coldict = lambda col: {
            "name": col.key,
            "type": col.type.__class__.__name__.upper(),
            "mode": "NULLABLE" if col.nullable else "REQUIRED",
        }
        return [coldict(col) for col in table_schema.get_children()]

    def save(self, payload: DataFrame):
        """
        Save payload to database table.

        Arguments:
            payload {DataFrame} -- the payload to be stored in db

        Raises:
            DaoError: if cannot save payload to db
        """
        try:
            # Always append date column
            # Todays date
            date = datetime.today().strftime("%Y-%m-%d")
            payload.insert(0, "date", date)
            pandas_gbq.to_gbq(
                payload, self._tablename, if_exists="append", table_schema=self._table,
            )
        except Exception as sql_exp:
            print(f"Exception in {self.tablename}")
            raise DaoError(sql_exp)

    def delete_all(self):
        """Delete all records in repository."""
        pandas_gbq.to_gbq(
            pd.DataFrame(),
            self.tablename,
            if_exists="replace",
            table_schema=self._table,
        )

    # pylint: disable=bad-continuation
    def query(
        self, query: str, chunksize: int, exact=False
    ) -> Generator[DataFrame, None, None]:
        """
        Query DAO repo and returns a generator of DataFrames with query results.

        Keyword Arguments:
            query {str} -- Query string
            chunksize {int} -- Number of rows of dataframe per chunk
            exact {bool} -- If false, query orders results and returns chunks

        Returns:
            Generator[DataFrame] -- Generator to iterate over DataFrame results.
        """
        if query is None:
            raise DaoError
        try:
            if exact:
                response_df = pandas_gbq.read_gbq(query)
                yield response_df
            if not exact:
                offset = 0
                # Remove semicolon if exists in original query to add ordering to query
                query = query.strip(";")
                while True:
                    add_query = (
                        f" ORDER BY {self._pkey} LIMIT {chunksize} OFFSET {offset};"
                    )
                    gbq_query = query + add_query
                    response_df = pandas_gbq.read_gbq(gbq_query)
                    if response_df.empty:
                        return
                    offset += chunksize
                    yield response_df
        except Exception as gbq_exp:
            raise DaoError(gbq_exp)



"""Module containing definitions of tables repository tables."""
from collections import namedtuple
from sqlalchemy import MetaData, Table, Column, Integer, Float, String

RepoProperties = namedtuple("RepoProperties", "tablename pkey table")
"""General properties of a given repository."""

CTD_PKEY = "digest"
"""CTD table primary key."""
CTD_TABLE_NAME = "ctd"
"""Name of ctd source table."""
CTD_DAO_TABLE = Table(
    CTD_TABLE_NAME,
    MetaData(),
    # No primary key allows duplicate records
    Column("digest", String, primary_key=False, nullable=False),
    Column("genesymbol", String),
    Column("geneid", Integer, nullable=False),
    Column("diseasename", String),
    Column("diseaseid", String, nullable=False),
    Column("pmids", String, nullable=False),
)
"""CTD DAO Repository Schema."""
CTD_PROPTY = RepoProperties(
    tablename=CTD_TABLE_NAME, pkey=CTD_PKEY, table=CTD_DAO_TABLE
)


PUBMED_TABLE_NAME = "pubmed"
"""PUBMED table primary key."""
PUBMED_PKEY = "pmid"
"""Name of pubmed source table."""
PUBMED_DAO_TABLE = Table(
    PUBMED_TABLE_NAME,
    MetaData(),
    # No primary key allows duplicate records
    Column("pmid", Integer, primary_key=False, nullable=False),
    Column("date_completed", String),
    Column("pub_model", String),
    Column("title", String),
    Column("iso_abbreviation", String),
    Column("article_title", String),
    Column("abstract", String),
    Column("authors", String),
    Column("language", String),
    Column("chemicals", String),
    Column("mesh_list", String),
    Column("issn", String),
    Column("issn_type", String),
    Column("citation_count", Integer),
    Column("citation_pmid", String),
)
"""PUBMED DAO Repository Schema."""
PUBMED_PROPTY = RepoProperties(
    tablename=PUBMED_TABLE_NAME, pkey=PUBMED_PKEY, table=PUBMED_DAO_TABLE
)

CLSFR_PKEY = "digest"
"""Classifier table primary key."""
CLSFR_TABLE_NAME = "classifier"
"""Classifier table name."""
CLSFR_TABLE_NAME = "classifier"
"""Name of geniepy classifier output table."""
CLSFR_DAO_TABLE = Table(
    CLSFR_TABLE_NAME,
    MetaData(),
    # No primary key allows duplicate records
    Column("digest", String, primary_key=False, nullable=False),
    Column("pub_score", Float, nullable=False),
    Column("ct_score", Float, nullable=False),
)
"""Classifier Output DAO Repository Schema."""
CLSFR_PROPTY = RepoProperties(
    tablename=CLSFR_TABLE_NAME, pkey=CLSFR_PKEY, table=CLSFR_DAO_TABLE
)


FEATURES_PKEY = "random_num"
"""Features table primary key."""
FEATURES_TABLE_NAME = "scoring_features"
"""Features table name."""
FEATURES_DAO_TABLE = None
"""Features Output DAO Repository Schema."""
FEATURES_PROPTY = RepoProperties(
    tablename=FEATURES_TABLE_NAME, pkey=FEATURES_PKEY, table=FEATURES_DAO_TABLE
)

SCORES_PKEY = "classifier_prob"
"""Scoring table primary key."""
SCORES_TABLE_NAME = "scores"
"""Scoring table name."""
SCORES_DAO_TABLE = None
"""Scoring Output DAO Repository Schema."""
SCORES_PROPTY = RepoProperties(
    tablename=SCORES_TABLE_NAME, pkey=SCORES_PKEY, table=SCORES_DAO_TABLE
)

PUBTATOR_GENE_PKEY = "GeneID"
"""Table primary key."""
PUBTATOR_GENE_TABLE_NAME = "gene2pubtator"
"""Table name."""
PUBTATOR_GENE_DAO_TABLE = Table(
    PUBTATOR_GENE_TABLE_NAME,
    MetaData(),
    # No primary key allows duplicate records
    Column("date", String, primary_key=False, nullable=False),
    Column("PMID", String),
    Column("GeneID", String),
)
"""Output DAO Repository Schema."""
PUBTATOR_GENE_PROPTY = RepoProperties(
    tablename=PUBTATOR_GENE_TABLE_NAME,
    pkey=PUBTATOR_GENE_PKEY,
    table=PUBTATOR_GENE_DAO_TABLE,
)

PUBTATOR_DISEASE_PKEY = "DiseaseID"
"""Table primary key."""
PUBTATOR_DISEASE_TABLE_NAME = "disease2pubtator"
"""Table name."""
PUBTATOR_DISEASE_DAO_TABLE = Table(
    PUBTATOR_DISEASE_TABLE_NAME,
    MetaData(),
    # No primary key allows duplicate records
    Column("date", String, primary_key=False, nullable=False),
    Column("PMID", String),
    Column("DiseaseID", String),
)
"""Output DAO Repository Schema."""
PUBTATOR_DISEASE_PROPTY = RepoProperties(
    tablename=PUBTATOR_DISEASE_TABLE_NAME,
    pkey=PUBTATOR_DISEASE_PKEY,
    table=PUBTATOR_DISEASE_DAO_TABLE,
)

SJR_PKEY = "Rank"
"""Table primary key."""
SJR_TABLE_NAME = "sjr"
"""Table name."""
SJR_DAO_TABLE = Table(
    SJR_TABLE_NAME,
    MetaData(),
    # No primary key allows duplicate records
    Column("date", String, primary_key=False, nullable=False),
    Column("Title", String),
    Column("SJR", String),
    Column("h_index", Integer),
)


"""Output DAO Repository Schema."""
SJR_PROPTY = RepoProperties(
    tablename=SJR_TABLE_NAME, pkey=SJR_PKEY, table=SJR_DAO_TABLE,
)



"""Test cron job functionalities."""
import pytest
import geniepy


@pytest.mark.slow_integration_test
class TestCronJob:
    """Test end-to-end functionalities."""

    def test_update_sjr(self):
        """Update dao table."""
        geniepy.update_sjr()

    def test_update_disease2pubtator(self):
        """Update dao table."""
        geniepy.update_disease2pubtator()

    def test_create_tables(self):
        """Update all tables."""
        geniepy.create_tables()

    def test_update_tables(self):
        """Update all tables."""
        geniepy.update_tables()

    def test_run_predictions(self):
        """Test calculating predictions."""
        geniepy.run_predictions()

    def test_sample(self):
        """Update all tables."""
        geniepy.sample_run()

    def test_run(self):
        """Test end-to-end run cron job function."""
        geniepy.run()



"""
Data Access Object Manager.

The DAO manager coordinates the DAOs from each one of the necessary sources to
generate the dataframe usef by the classifiers to calculate a prediction score.
"""
from typing import Generator
import pandas as pd
import geniepy.datamgmt.daos as daos
from multiprocessing import Process
import geniepy.config as config
import geniepy.datamgmt.tables as gt
import geniepy.datamgmt.repositories as dr


class DaoManager:
    """
    Implementation of Data Access Object Manager.

    The DAO manager is intended to utilize different DAOs to generate
    dataframe records used by the classifiers.

    At start of day the download method should be called to scrape data from online
    sources on each DAO and generate appropriate tables.

    The DAO manager should then be called to generate the records that can be fed into
    the classifiers.
    """

    __slots__ = [
        "_sjr_dao",
        "_pubtator_disease_dao",
        "_pubtator_gene_dao",
        "_pubmed_dao",
        "_features",
        "_scores",
    ]

    def create_repos(self):
        """Configure and create scoring repositories."""
        credentials = config.get_credentials()
        projname = config.get_projname()
        dataset = config.get_dataset("scoring")
        features_repo = dr.GbqRepository(
            projname, gt.FEATURES_PROPTY, dataset, credentials
        )
        scores_repo = dr.GbqRepository(projname, gt.SCORES_PROPTY, dataset, credentials)
        return features_repo, scores_repo

    # pylint: disable=bad-continuation
    def __init__(
        self,
        sjr_dao: daos.SjrDao,
        pubtator_disease_dao: daos.PubtatorDiseaseParser,
        pubtator_gene_dao: daos.PubtatorGeneParser,
        pubmed_dao: daos.PubMedDao,
    ):
        """Initializa DAO mgr with corresponding DAO children."""
        self._sjr_dao = sjr_dao
        self._pubtator_disease_dao = pubtator_disease_dao
        self._pubtator_gene_dao = pubtator_gene_dao
        self._pubmed_dao = pubmed_dao
        self._features, self._scores = self.create_repos()

    def download(self, chunksize: int, **kwargs):
        """Download (scrapes) data for DAOs and creates internal tables."""
        # Fire off scrapers async
        psjr = Process(target=self._sjr_dao.download, args=(chunksize,), kwargs=kwargs)
        ppubtatordisease = Process(
            target=self._pubtator_disease_dao.download, args=(chunksize,), kwargs=kwargs
        )
        ppubtatorgene = Process(
            target=self._pubtator_gene_dao.download, args=(chunksize,), kwargs=kwargs
        )
        ppubmed = Process(
            target=self._pubmed_dao.download, args=(chunksize,), kwargs=kwargs
        )
        psjr.start()
        ppubtatordisease.start()
        ppubtatorgene.start()
        ppubmed.start()

    def _get_pubmeds_df(self, pmids: str):
        """
        Get pubmed dao dataframes.

        Arguments:
            pubmeds {str} -- pipe delimtied string with pmids
            chunksize {int} -- limits max chunksize internally to limit memory usage

        Returns: Dataframe with pubmed dao dataframe.
        """
        pmids = pmids.split("|")
        pubmed_df = pd.DataFrame()
        for pmid in pmids:
            try:
                pmid_query = self._pubmed_dao.query_pkey(int(pmid))
                # Only care about 1 pmid entry (table shouldn't have duplicates)
                pmid_df = next(self._pubmed_dao.query(pmid_query, 1))
                pubmed_df = pubmed_df.append(pmid_df, ignore_index=True)
            except StopIteration:  # pragma: no cover
                # TODO log instead of print
                print(f"PMID: {pmid} not found!")
        return pubmed_df

    def get_max_feature(self, chunksize):
        """Retrieve the max addressable record in features table."""
        count_query = self._features.query_all.replace("*", "MAX(random_num)")
        max_df = next(self._features.query(count_query, 1, exact=True))
        # Make sure go past max to include all numbers in range
        return int(max_df.iloc[0][0]) + chunksize

    def get_features(self, offset: int, chunksize: int) -> pd.DataFrame:
        """
        Generate the dataframe records for classifiers.

        This function should be called after the DAO tables have been created through
        the download function.

        Arguments:
            offset {int} -- The offset of records to be fetched
            chunksize {int} -- The number of records to be returned


        Returns:
            A dataframe containing records from features table
        """
        gen_query = (
            lambda offset: self._features.query_all
            + f" WHERE random_num BETWEEN {offset} AND {offset + chunksize};"
        )
        query_str = gen_query(offset)
        print(query_str)
        record_df = next(self._features.query(query_str, chunksize, exact=True))
        return record_df

    def save_predictions(self, predictions: pd.DataFrame):
        """
        Save computed predictions and supporting data into output tables.

        Arguments:
            records {DataFrame} -- [description]
        """
        self._scores.save(predictions)



"""Module with data used in tests."""
# flake8: noqa
# pylint: skip-file
import pandas as pd


PUBMED_INVALID_SCHEMA = [
    pd.DataFrame(
        {
            # Missing required column
            "date_completed": ["--"],
            "pub_model": ["Print-Electronic"],
            "title": ["Biochemia medica"],
            "iso_abbreviation": ["Biochem Med (Zagreb)"],
            "article_title": [
                "High anion gap metabolic acidosis caused by D-lactate: mind the time of blood collection."  # NOQA
            ],
            "abstract": [
                "D-lactic acidosis is an uncommon cause of high anion gap acidosis."
            ],
            "authors": ["Weemaes, Matthias, Hiele, Martin, Vermeersch, Pieter"],
            "language": ["eng"],
            "chemicals": [""],
            "mesh_list": [""],
        }
    ),
    pd.DataFrame(
        {
            "pmid": 31839729,
            "not column": [""],  # Invalid Column
            "pub_model": ["Print-Electronic"],
            "title": ["Biochemia medica"],
            "iso_abbreviation": ["Biochem Med (Zagreb)"],
            "article_title": [
                "Unexpected abnormal coagulation test results in a 2-year-old child: A case report."
            ],
            "abstract": [
                "Rejection of the sample with repeated blood withdrawal is always an unwanted consequence of sample nonconformity and preanalytical errors, especially in the most vulnerable population - children. Here is presented a case with unexpected abnormal coagulation test results in a 2-year-old child with no previously documented coagulation disorder. Child is planned for tympanostomy tubes removal under the anaesthesia driven procedure, and preoperative coagulation tests revealed prolonged prothrombin time, activated partial thromboplastin time and thrombin time, with fibrinogen and antithrombin within reference intervals. From the anamnestic and clinical data, congenital coagulation disorder was excluded, and with further investigation, sample mismatch, clot presence and accidental ingestion of oral anticoagulant, heparin contamination or vitamin K deficiency were excluded too. Due to suspected EDTA carryover during blood sampling another sample was taken the same day and all tests were performed again. The results for all tests were within reference intervals confirming EDTA effect on falsely prolongation of the coagulation times in the first sample. This case can serve as alert to avoid unnecessary loss in terms of blood withdrawal repetitions and discomfort of the patients and their relatives, tests repeating, prolonging medical procedures, and probably delaying diagnosis or proper medical treatment. It is the responsibility of the laboratory specialists to continuously educate laboratory staff and other phlebotomists on the correct blood collection as well as on its importance for the patient's safety."
            ],
            "authors": [
                "Banković Radovanović, Patricija, Živković Mikulčić, Tanja, Simović Medica, Jasmina"
            ],
            "language": ["eng"],
            "chemicals": [""],
            "mesh_list": [""],
        }
    ),
]


PUBMED_INVALID_DF = [
    pd.DataFrame(
        {
            "pmid": ["31839728"],  # pmid as string
            "date_completed": ["--"],
            "pub_model": ["Print-Electronic"],
            "title": ["Biochemia medica"],
            "iso_abbreviation": ["Biochem Med (Zagreb)"],
            "article_title": [
                "High anion gap metabolic acidosis caused by D-lactate: mind the time of blood collection."
            ],
            "abstract": [
                "D-lactic acidosis is an uncommon cause of high anion gap acidosis."
            ],
            "authors": ["Weemaes, Matthias, Hiele, Martin, Vermeersch, Pieter"],
            "language": ["eng"],
            "chemicals": [""],
            "mesh_list": [""],
        }
    ),
    pd.DataFrame(
        {
            "pmid": 31839729,
            # missing column
            "pub_model": ["Print-Electronic"],
            "title": ["Biochemia medica"],
            "iso_abbreviation": ["Biochem Med (Zagreb)"],
            "article_title": [
                "Unexpected abnormal coagulation test results in a 2-year-old child: A case report."
            ],
            "abstract": [
                "Rejection of the sample with repeated blood withdrawal is always an unwanted consequence of sample nonconformity and preanalytical errors, especially in the most vulnerable population - children. Here is presented a case with unexpected abnormal coagulation test results in a 2-year-old child with no previously documented coagulation disorder. Child is planned for tympanostomy tubes removal under the anaesthesia driven procedure, and preoperative coagulation tests revealed prolonged prothrombin time, activated partial thromboplastin time and thrombin time, with fibrinogen and antithrombin within reference intervals. From the anamnestic and clinical data, congenital coagulation disorder was excluded, and with further investigation, sample mismatch, clot presence and accidental ingestion of oral anticoagulant, heparin contamination or vitamin K deficiency were excluded too. Due to suspected EDTA carryover during blood sampling another sample was taken the same day and all tests were performed again. The results for all tests were within reference intervals confirming EDTA effect on falsely prolongation of the coagulation times in the first sample. This case can serve as alert to avoid unnecessary loss in terms of blood withdrawal repetitions and discomfort of the patients and their relatives, tests repeating, prolonging medical procedures, and probably delaying diagnosis or proper medical treatment. It is the responsibility of the laboratory specialists to continuously educate laboratory staff and other phlebotomists on the correct blood collection as well as on its importance for the patient's safety."
            ],
            "authors": [
                "Banković Radovanović, Patricija, Živković Mikulčić, Tanja, Simović Medica, Jasmina"
            ],
            "language": ["eng"],
            "chemicals": [""],
            "mesh_list": [""],
        }
    ),
    pd.DataFrame(
        {
            "pmid": 31839730,
            "date_completed": ["--"],
            "pub_model": ["Print-Electronic"],
            "title": ["Narrative inquiry : NI"],
            "iso_abbreviation": ["Narrat Inq"],
            "article_title": [
                "Narrative Assessments with First Grade Spanish-English Emergent Bilinguals: Spontaneous versus Retell Conditions."
            ],
            "abstract": [
                "This study used qualitative analyses to investigate similarities and differences in narrative production across two task conditions for four first grade Spanish-English emergent bilingual children. Task conditions were spontaneous story generation and retelling using the same story. Spanish stories from two children were compared on the basis of similarity in vocabulary, while English stories from two children were compared on the basis of similarity in overall discourse skills. Results show that when the total number of words used was similar across English narratives, the retell included more different words and higher quality story structure than the spontaneous story. When overall discourse scores in the Spanish examples were similar, the spontaneous story required more words than the retell, but also included more central events and greater detail. Yet, the retell included more advanced narrative components. This study contributes to our understanding of narrative skills in young Spanish-English bilinguals across task conditions."
            ],
            "authors": ["Lucero, Audrey, Uchikoshi, Yuuko"],
            "language": ["eng"],
            "chemicals": [""],
            "mesh_list": [""],
            "other col": [""],  # Extra column
        }
    ),
    pd.DataFrame(
        {
            "pmid": 31839731,
            "date_completed": ["--"],
            "pub_model": ["Print"],
            "title": ["Acta ortopedica brasileira"],
            "iso_abbreviation": ["Acta Ortop Bras"],
            "article_title": [
                "CHRONIC MONTEGGIA FRACTURE-DISLOCATION IN CHILDREN SURGICAL STRATEGY AND RESULTS."
            ],
            "abstract": [
                "To report surgical techniques and results in the treatment of chronic Monteggia fracture-dislocation in children."
            ],
            "authors": [
                "Soni, Jamil Faissal, Valenza, Weverley Rubele, Matsunaga, Carolina Umeta, Costa, Anna Carolina Pavelec, Faria, Fernando Ferraz"
            ],
            "language": ["eng"],
            # Missing column
            "mesh_list": [""],
        }
    ),
    pd.DataFrame(
        {
            "pmid": 31839732,
            "date_completed": ["--"],
            "pub_model": ["Print"],
            "title": ["Acta ortopedica brasileira"],
            # Missing Column
            # Missing Column
            "abstract": [
                "To evaluate the efficacy of platelet-rich plasma (PRP) and tranexamic acid (TXA) applied in total knee arthroplasty."
            ],
            "authors": [
                "Guerreiro, João Paulo Fernandes, Lima, Diogenes Rodrigues, Bordignon, Glaucia, Danieli, Marcus Vinicius, Queiroz, Alexandre Oliveira, Cataneo, Daniele Cristina"
            ],
            "language": ["eng"],
            "chemicals": [""],
            "mesh_list": [""],
        }
    ),
]
"""Array of invalid PubMed dataframes."""


PUBMED_VALID_DF = [
    pd.DataFrame(
        {
            "pmid": 31839728,
            "date_completed": ["--"],
            "pub_model": ["Print-Electronic"],
            "title": ["Biochemia medica"],
            "iso_abbreviation": ["Biochem Med (Zagreb)"],
            "article_title": [
                "High anion gap metabolic acidosis caused by D-lactate: mind the time of blood collection."
            ],
            "abstract": [
                "D-lactic acidosis is an uncommon cause of high anion gap acidosis."
            ],
            "authors": ["Weemaes, Matthias, Hiele, Martin, Vermeersch, Pieter"],
            "language": ["eng"],
            "chemicals": [""],
            "mesh_list": [""],
            "issn": ["1234-5678"],
            "issn_type": ["Print"], 
            "citation_count": 0,
            "citation_pmid": [""],
        }
    ),
    pd.DataFrame(
        {
            "pmid": 31839729,
            "date_completed": ["--"],
            "pub_model": ["Print-Electronic"],
            "title": ["Biochemia medica"],
            "iso_abbreviation": ["Biochem Med (Zagreb)"],
            "article_title": [
                "Unexpected abnormal coagulation test results in a 2-year-old child: A case report."
            ],
            "abstract": [
                "Rejection of the sample with repeated blood withdrawal is always an unwanted consequence of sample nonconformity and preanalytical errors, especially in the most vulnerable population - children. Here is presented a case with unexpected abnormal coagulation test results in a 2-year-old child with no previously documented coagulation disorder. Child is planned for tympanostomy tubes removal under the anaesthesia driven procedure, and preoperative coagulation tests revealed prolonged prothrombin time, activated partial thromboplastin time and thrombin time, with fibrinogen and antithrombin within reference intervals. From the anamnestic and clinical data, congenital coagulation disorder was excluded, and with further investigation, sample mismatch, clot presence and accidental ingestion of oral anticoagulant, heparin contamination or vitamin K deficiency were excluded too. Due to suspected EDTA carryover during blood sampling another sample was taken the same day and all tests were performed again. The results for all tests were within reference intervals confirming EDTA effect on falsely prolongation of the coagulation times in the first sample. This case can serve as alert to avoid unnecessary loss in terms of blood withdrawal repetitions and discomfort of the patients and their relatives, tests repeating, prolonging medical procedures, and probably delaying diagnosis or proper medical treatment. It is the responsibility of the laboratory specialists to continuously educate laboratory staff and other phlebotomists on the correct blood collection as well as on its importance for the patient's safety."
            ],
            "authors": [
                "Banković Radovanović, Patricija, Živković Mikulčić, Tanja, Simović Medica, Jasmina"
            ],
            "language": ["eng"],
            "chemicals": [""],
            "mesh_list": [""],
            "issn": ["1234-5678"],
            "issn_type": ["Print"], 
            "citation_count": 0,
            "citation_pmid": [""],            
        }
    ),
    pd.DataFrame(
        {
            "pmid": 31839730,
            "date_completed": ["--"],
            "pub_model": ["Print-Electronic"],
            "title": ["Narrative inquiry : NI"],
            "iso_abbreviation": ["Narrat Inq"],
            "article_title": [
                "Narrative Assessments with First Grade Spanish-English Emergent Bilinguals: Spontaneous versus Retell Conditions."
            ],
            "abstract": [
                "This study used qualitative analyses to investigate similarities and differences in narrative production across two task conditions for four first grade Spanish-English emergent bilingual children. Task conditions were spontaneous story generation and retelling using the same story. Spanish stories from two children were compared on the basis of similarity in vocabulary, while English stories from two children were compared on the basis of similarity in overall discourse skills. Results show that when the total number of words used was similar across English narratives, the retell included more different words and higher quality story structure than the spontaneous story. When overall discourse scores in the Spanish examples were similar, the spontaneous story required more words than the retell, but also included more central events and greater detail. Yet, the retell included more advanced narrative components. This study contributes to our understanding of narrative skills in young Spanish-English bilinguals across task conditions."
            ],
            "authors": ["Lucero, Audrey, Uchikoshi, Yuuko"],
            "language": ["eng"],
            "chemicals": [""],
            "mesh_list": [""],
            "issn": ["1234-5678"],
            "issn_type": ["Print"], 
            "citation_count": 0,
            "citation_pmid": [""],            
        }
    ),
    pd.DataFrame(
        {
            "pmid": 31839731,
            "date_completed": ["--"],
            "pub_model": ["Print"],
            "title": ["Acta ortopedica brasileira"],
            "iso_abbreviation": ["Acta Ortop Bras"],
            "article_title": [
                "CHRONIC MONTEGGIA FRACTURE-DISLOCATION IN CHILDREN SURGICAL STRATEGY AND RESULTS."
            ],
            "abstract": [
                "To report surgical techniques and results in the treatment of chronic Monteggia fracture-dislocation in children."
            ],
            "authors": [
                "Soni, Jamil Faissal, Valenza, Weverley Rubele, Matsunaga, Carolina Umeta, Costa, Anna Carolina Pavelec, Faria, Fernando Ferraz"
            ],
            "language": ["eng"],
            "chemicals": [""],
            "mesh_list": [""],
            "issn": ["1234-5678"],
            "issn_type": ["Print"], 
            "citation_count": 0,
            "citation_pmid": [""],            
        }
    ),
    pd.DataFrame(
        {
            "pmid": 31839732,
            "date_completed": ["--"],
            "pub_model": ["Print"],
            "title": ["Acta ortopedica brasileira"],
            "iso_abbreviation": ["Acta Ortop Bras"],
            "article_title": [
                "PLATELET-RICH PLASMA (PRP) AND TRANEXAMIC ACID (TXA) APPLIED IN TOTAL KNEE ARTHROPLASTY."
            ],
            "abstract": [
                "To evaluate the efficacy of platelet-rich plasma (PRP) and tranexamic acid (TXA) applied in total knee arthroplasty."
            ],
            "authors": [
                "Guerreiro, João Paulo Fernandes, Lima, Diogenes Rodrigues, Bordignon, Glaucia, Danieli, Marcus Vinicius, Queiroz, Alexandre Oliveira, Cataneo, Daniele Cristina"
            ],
            "language": ["eng"],
            "chemicals": [""],
            "mesh_list": [""],
            "issn": ["1234-5678"],
            "issn_type": ["Print"], 
            "citation_count": 0,
            "citation_pmid": [""],            
        }
    ),
]
"""Array of valid PubMed dataframes."""

CTD_INVALID_SCHEMA = [
    pd.DataFrame(
        {
            # Missing Digest required field
            "genesymbol": ["11-BETA-HSD3"],
            "geneid": [100174880],
            "diseasename": ["Abnormalities, Drug-Induced"],
            "diseaseid": ["D000014"],
            "pmids": ["22659286"],
        }
    ),
    pd.DataFrame(
        {
            "OtherField": [22659286],  # Non-existent schema column
            "digest": [22659286],
            "genesymbol": ["11-BETA-HSD3"],
            "geneid": [100174880],
            "diseasename": ["Abnormalities, Drug-Induced"],
            "diseaseid": ["D000014"],
            "pmids": ["22659286"],
        }
    ),
    pd.DataFrame(
        {
            "digest": [22659286],
            "genesymbol": ["11-BETA-HSD3"],
            # Missing required GeneID
            "diseasename": ["Abnormalities, Drug-Induced"],
            "diseaseid": ["D000014"],
            "pmids": ["22659286"],
        }
    ),
    pd.DataFrame(
        {
            "digest": [22659286],
            "genesymbol": ["11-BETA-HSD3"],
            "geneid": [100174880],
            "diseasename": ["Abnormalities, Drug-Induced"],
            # Missing required DiseaseID
            "pmids": ["22659286"],
        }
    ),
    pd.DataFrame(
        {
            "digest": [22659286],
            "genesymbol": ["11-BETA-HSD3"],
            "geneid": [100174880],
            "diseasename": ["Abnormalities, Drug-Induced"],
            "diseaseid": ["D000014"],
            # Missing required PubMed
        }
    ),
]
"""Invalid SCHEMA record."""

CTD_INVALID_DF = [
    None,
    pd.DataFrame({"invalid": [1, 2]}),
    pd.DataFrame(
        {
            "digest": [22659286],
            "genesymbol": ["11-BETA-HSD3"],
            "geneid": ["A100174880"],  # Should be Int
            "diseasename": ["Abnormalities, Drug-Induced"],
            "diseaseid": ["D000014"],
            "pmids": ["22659286"],
        }
    ),
    pd.DataFrame(
        {
            "digest": [22659286],
            "genesymbol": ["11-BETA-HSD3"],
            "geneid": [100174880.0],  # Should be Int
            "diseasename": ["Abnormalities, Drug-Induced"],
            "diseaseid": ["D000014"],
            "pmids": ["22659286"],
        }
    ),
    pd.DataFrame(
        {
            "digest": [22659286],
            "genesymbol": ["11-BETA-HSD3"],
            "geneid": [100174880],
            "diseasename": ["Abnormalities, Drug-Induced"],
            "diseaseid": ["MESH:D000014"],  # Should not have "MESH:"
            "pmids": ["22659286"],
        }
    ),
] + CTD_INVALID_SCHEMA
"""Array of invalid CTD DataFrames because violate parser rules."""


CTD_VALID_DF = [
    pd.DataFrame(
        {
            "digest": [
                "b3834d9281286247e377c5700e9689c3660412df24fa0a4921c6e3c213d616aa"
            ],
            "genesymbol": ["11-BETA-HSD3"],
            "geneid": [10174880],
            "diseasename": ["Abnormalities, Drug-Induced"],
            "diseaseid": ["D000014"],
            "pmids": ["22659286"],
        }
    ),
    pd.DataFrame(
        {
            "digest": [
                "e96cc1eb2423dad1fd6f4f341574fbb7fff0479a3339c9e2e4f814f2d970e3f00"
            ],
            "genesymbol": ["1-SF3"],
            "geneid": [1000494280],
            "diseasename": ["Infant Death"],
            "diseaseid": ["D0660884"],
            "pmids": ["283930756"],
        }
    ),
    pd.DataFrame(
        {
            "digest": [
                "e120bfecd61ee146bb3a4c61f8dbb93754b1db25f62aba505872ef5568dd155b"
            ],
            "genesymbol": ["11-BETA-HSD3"],
            "geneid": [100174880],
            "diseasename": ["Abnormalities, Drug-Induced"],
            "diseaseid": ["D000014"],
            "pmids": ["22659286"],
        }
    ),
    pd.DataFrame(
        {
            "digest": [
                "f847dcfeaaae8fefa96e1f6e97dd5998ffe07365d1254badae927d1769a2eebc"
            ],
            "genesymbol": ["A1BG3"],
            "geneid": [1],
            "diseasename": ["Muscle Weaknessd"],
            "diseaseid": ["D0189084"],
            "pmids": ["3515563|54800|62135|63766|6511338|7995496"],
        }
    ),
]
"""Array of valid CTD DataFrames."""

CLSFR_VALID_DF = [
    pd.DataFrame(
        {
            "digest": [
                "b3834d9281286247e377c5700e9689c3660412df24fa0a4921c6e3c213d616aa"
            ],
            "pub_score": [0.8],
            "ct_score": [0.7],
        }
    ),
    pd.DataFrame(
        {
            "digest": [
                "e96cc1eb2423dad1fd6f4f341574fbb7fff0479a3339c9e2e4f814f2d970e3f00"
            ],
            "pub_score": [0.8],
            "ct_score": [0.7],
        }
    ),
    pd.DataFrame(
        {
            "digest": [
                "e96cc1eb2423dad1fd6f4f341574fbb7fff0479a3339c9e2e4f814f2d970e3f00"
            ],
            "pub_score": [0.8],
            "ct_score": [0.7],
        }
    ),
]
"""Array of valid classifier dataframes."""


CLSFR_INVALID_DF = [
    pd.DataFrame(
        {
            "digest": [
                "b3834d9281286247e377c5700e9689c3660412df24fa0a4921c6e3c213d616aa"
            ],
            "pub_score": ["1.2"],  # GeneID should be float
            "ct_score": [0.7],
        }
    ),
    pd.DataFrame(
        {
            "digest": [
                "b3834d9281286247e377c5700e9689c3660412df24fa0a4921c6e3c213d616aa"
            ],
            # Missing pub_score
            "ct_score": [0.7],
        }
    ),
    pd.DataFrame(
        {
            "digest": [
                "e96cc1eb2423dad1fd6f4f341574fbb7fff0479a3339c9e2e4f814f2d970e3f00"
            ],
            "pub_score": [0.8],
            "ct_score": ["0.75"],  # Should be float
        }
    ),
]
"""Array of invalid classifier dataframes."""


CLSFR_INVALID_SCHEMA = [
    pd.DataFrame(
        {
            "digest": [
                "b3834d9281286247e377c5700e9689c3660412df24fa0a4921c6e3c213d616aa"
            ],
            # Missing GeneID
            "ct_score": [0.7],
        }
    ),
    pd.DataFrame(
        {
            "digest": [
                "e96cc1eb2423dad1fd6f4f341574fbb7fff0479a3339c9e2e4f814f2d970e3f00"
            ],
            "pub_score": [0.8],
            "ct_score": [0.7],
            "newcol": "newcol",
        }
    ),
    pd.DataFrame(
        {
            # missing digest
            "pub_score": [0.8],
            "ct_score": [0.7],
        }
    ),
    pd.DataFrame(
        {
            "digest": [
                "e96cc1eb2423dad1fd6f4f341574fbb7fff0479a3339c9e2e4f814f2d970e3f00"
            ],
            "pub_score": [0.8],
            # Missing DiseaseID
        }
    ),
]
"""Array of valid classifier SCHEMA table schema."""



"""
Classification Manager Module.

The Classifier Manager coordinates the received data to run it through each classifier,
generate the appropriate predictions and return the classifier dataframe according to
the ClassifierParser schema.
"""
import pandas as pd
from geniepy.classmgmt.classifiers import BaseClassifier
from geniepy.errors import ClassifierError


class ClassificationMgr:
    """
    Classification Manager Implementation.

    The classification manager is handles multiple classifiers by calling their
    respective predict methods on the data and returning an assembeld dataframe.
    """

    __slots__ = ["_classifiers"]

    def __init__(self, classifiers: [BaseClassifier]):
        """
        Initialize classification manager and train classifiers.

        Arguments:
            classifiers {[type]} -- The list of classifiers to be managed.
        """
        self._classifiers: [BaseClassifier] = classifiers
        # TODO Train or restore classifiers.
        for classifier in self._classifiers:
            # TODO Replace to load correct model
            classifier.load()

    def predict(self, records: pd.DataFrame):
        """
        Run prediction on input data through all classifiers.

        Arguments:
            records {pd.DataFrame} -- The records to be predicted as a dataframe.
        """
        # Generate prediction dataframe keys
        try:
            prediction_df = pd.DataFrame()
            prediction_df["digest"] = records["digest"]
            for classifier in self._classifiers:
                prediction_df[classifier.name] = records.apply(
                    classifier.predict, axis=1
                )
            return prediction_df
        except Exception as exp:
            raise ClassifierError("Unable to predict records: " + str(exp))



"""Module containing helper functions for running genie scripts."""
import gzip
import shutil


def decompress_gz(file_path: str, output_file: str):
    """
    Unzip articles sets and create xml file.

    Arguments:
        file_path {str} -- absolute path gz file
        output_file {str} -- absolute path to created decompressed file
    """
    with gzip.open(file_path, "rb") as f_in:
        with open(output_file, "wb") as f_out:
            shutil.copyfileobj(f_in, f_out)



from csv import reader
import requests
import xmltodict
import urllib.parse
import re
from datetime import datetime
import csv
import concurrent.futures

# Constants
base_url = "https://www.ncbi.nlm.nih.gov/pmc"
converter_path = "/utils/idconv/v1.0/?tool=my_tool&email=my_email@example.com&format=json"
retrieval_path = "/oai/oai.cgi?verb=GetRecord&metadataPrefix=pmc_fm&identifier=oai:pubmedcentral.nih.gov:"
api_key = "9ab4b04dcabcab740d1a297e6f3f54aa1e09"
EMAIL_REGEX = re.compile(r"[^@]+@[^@]+\.[^@]+")
dateTimeObj = datetime.now().__str__()
batch_size = 10
csv_file = "pubmed_id.csv"


def find(term, dictionary):
    for k, v in dictionary.items():
        if k == term:
            yield dictionary
        elif isinstance(v, dict):
            for result in find(term, v):
                yield result
        elif isinstance(v, list):
            for d in v:
                if isinstance(d, dict):
                    for result in find(term, d):
                        yield result


def email_lookup(pubmed_ids):
    out_csv_file = "author_info_emails.csv"
    email_dict = dict()

    with open(out_csv_file, 'a') as csvfile:
        for pubmed_id in pubmed_ids:
            params_converter = {
                'ids': pubmed_id,
                'api_key': api_key
            }

            convert_request = requests.get(url=base_url + converter_path, params=params_converter)
            converted_data = convert_request.json()

            if list(find("errmsg", converted_data)):
                # ID not found in Pub med
                continue
            else:
                pmc_id = converted_data['records'][0]['pmcid']
                decoded_url = urllib.parse.unquote(base_url + retrieval_path + pmc_id[3:])

                retrieval_request = requests.get(url=decoded_url)

                retrieval_data = xmltodict.parse(retrieval_request.content)

                email_obj = list(find("email", retrieval_data))
                email_set = set()

                # Email could exist anywhere or everywhere up till 4th nested level
                # This is being done to get the max emails data extraction from the pubmed ids
                if email_obj:
                    if isinstance(email_obj, list) | isinstance(email_obj, dict):
                        for entry in email_obj:
                            if isinstance(entry, dict):
                                for value in entry.values():
                                    if isinstance(value, dict):
                                        for v1 in value.values():
                                            if EMAIL_REGEX.match(str(v1)):
                                                if str(v1).startswith('mailto:'):
                                                    v1 = str(v1)[7:]
                                                email_set.add(str(v1))
                                    if isinstance(value, list):
                                        for v2 in value:
                                            if isinstance(v2, dict):
                                                for v3 in v2.values():
                                                    if EMAIL_REGEX.match(str(v3)):
                                                        if str(v3).startswith('mailto:'):
                                                            v3 = str(v3)[7:]
                                                        email_set.add(str(v3))
                                            else:
                                                if EMAIL_REGEX.match(str(v2)):
                                                    email_set.add(str(v2))
                                    elif isinstance(value, dict):
                                        for v4 in value.values():
                                            if EMAIL_REGEX.match(str(v4)):
                                                if str(v4).startswith('mailto:'):
                                                    v4 = str(v4)[7:]
                                                email_set.add(str(v4))
                                    else:
                                        if EMAIL_REGEX.match(str(value)):
                                            email_set.add(str(value))
                            else:
                                if EMAIL_REGEX.match(str(entry)):
                                    email_set.add(str(entry))

                    # print(str(pubmed_id) + ", " + str(email_set))

                    for address in email_set:
                        email_dict.update({str(address): str(pubmed_id)})

                    csv_file_writer = csv.writer(csvfile)
                    for key, value in email_dict.items():
                        csv_file_writer.writerow([key, str(value)])
                    email_dict.clear()
        csvfile.close()


def main():
    with open(csv_file, 'r') as read_obj:
        csv_reader = reader(read_obj)
        pubmed_ids = list(csv_reader)
    with concurrent.futures.ProcessPoolExecutor(max_workers=8) as executor:
        
        result = executor.map(email_lookup, pubmed_ids)
        print(result)


if __name__ == '__main__':
    main()



"""Module to test data access object repositories."""
import pytest
import tests.testdata as td
from geniepy.datamgmt.repositories import BaseRepository, SqlRepository
from geniepy.datamgmt.tables import PUBMED_PROPTY
from geniepy.errors import DaoError
from tests.resources.mock import TEST_CHUNKSIZE

VALID_DF = td.PUBMED_VALID_DF
INVALID_SCHEMA = td.PUBMED_INVALID_SCHEMA


class TestSqlPubMedRepository:
    """PyTest repository test class."""

    repo: BaseRepository = SqlRepository("sqlite://", PUBMED_PROPTY)

    def test_constructor(self):
        """Ensure scraper obj constructed successfully."""
        assert self.repo is not None

    def test_query_all(self):
        """Test gen query all str."""
        expected = "SELECT * FROM pubmed"
        actual = self.repo.query_all
        assert actual == expected

    def test_query_pkey(self):
        """Test gen query all str."""
        pmid = 1
        expected = "SELECT * FROM pubmed WHERE pmid=1"
        actual = self.repo.query_pkey(pmid)
        assert actual == expected

    @pytest.mark.parametrize("payload", INVALID_SCHEMA)
    def test_save_invalid_df(self, payload):
        """Test save invalid dataframe to dao's DAO."""
        with pytest.raises(DaoError):
            self.repo.save(payload)

    @pytest.mark.parametrize("payload", VALID_DF)
    def test_save_valid_df(self, payload):
        """Attempt to save dataframe with valid schema."""
        self.repo.save(payload)  # Don't expect to return anything

    @pytest.mark.parametrize("payload", VALID_DF)
    def test_query(self, payload):
        """Query valid record."""
        # Start with empty table
        self.repo.delete_all()
        # Try to create records in db for test if don't exist
        try:
            self.repo.save(payload)
        except DaoError:
            pass
        # Attempt to retrieve record
        pmid = payload.pmid[0]
        query_str = self.repo.query_pkey(pmid)
        generator = self.repo.query(query_str, TEST_CHUNKSIZE)
        chunk = next(generator)
        assert chunk.equals(payload)

    def test_query_non_existent(self):
        """Query non-existent record should return empty."""
        # Attempt to retrieve record
        pmid = 0
        query_str = self.repo.query_pkey(pmid)
        generator = self.repo.query(query_str, TEST_CHUNKSIZE)
        # Make sure generator doesn't return anything since no matching records
        with pytest.raises(StopIteration):
            next(generator)

    @pytest.mark.parametrize("chunksize", [1, 2, 3, 4])
    def test_generator_chunk(self, chunksize):
        """Query all by chunk."""
        # Start with empty table
        self.repo.delete_all()
        # Try to fill database, in case is empty
        for record in VALID_DF:
            try:
                self.repo.save(record)
            except DaoError:
                pass
        # Get all records in database
        generator = self.repo.query(self.repo.query_all, chunksize)
        # Make sure number generator provides df of chunksize each iteration
        result_df = next(generator)
        assert result_df.pmid.count() == chunksize

    def test_delete_all(self):
        """Test delete all records from repository."""
        # Try to fill database, in case is empty
        for record in VALID_DF:
            try:
                self.repo.save(record)
            except DaoError:
                pass
        # Delete all records
        self.repo.delete_all()
        # Make sure no records left
        generator = self.repo.query(self.repo.query_all, TEST_CHUNKSIZE)
        # generator shouldn't return anything since no records in database
        with pytest.raises(StopIteration):
            next(generator)
        # Test saving and reading from table again, make sure still functional
        self.repo.save(VALID_DF[0])
        generator = self.repo.query(self.repo.query_all, TEST_CHUNKSIZE)
        # Generator should return value
        next(generator)



"""Module to test online sources parsers."""
import pytest
from geniepy.datamgmt.parsers import BaseParser, ClassifierParser
import tests.testdata as td
from tests.resources.mock import TEST_CHUNKSIZE


VALID_DF = td.CLSFR_VALID_DF
INVALID_DF = td.CLSFR_INVALID_DF


class TestClassifierParser:
    """Pytest CTD Parser class."""

    parser: BaseParser = ClassifierParser()

    def test_constructor(self):
        """Ensure scraper obj constructed successfully."""
        assert self.parser is not None

    @pytest.mark.parametrize("payload", INVALID_DF)
    def test_invalid_payload(self, payload):
        """Test invalid dataframe."""
        # Should return list with errors
        assert self.parser.validate(payload)

    @pytest.mark.parametrize("payload", VALID_DF)
    def test_valid_payload(self, payload):
        """Test valid dataframe."""
        # Should return empty list
        assert not self.parser.validate(payload)

    def test_parse_not_impl(self):
        """Parse method should not be implemented in classifier parser."""
        with pytest.raises(NotImplementedError):
            self.parser.parse(None)

    def test_fetch_not_impl(self):
        """Fetch method should not be implemented in classifier parser."""
        with pytest.raises(NotImplementedError):
            self.parser.fetch(TEST_CHUNKSIZE)



"""
Handle pubmed historical data.

Parse the pubmed baseline article sets and generate jsonl PubMedArticles to be
used by classifier.

The script expects path to folder containing pubmed baseline .xml.gz files,
the path to output directory where generated files should be stored, and
number of concurrent processes to be used [1, 16].
"""
# pylint: disable=wrong-import-order, unused-import
import geniebootsrap  # noqa: F401
import logging
import sys
import os
import gzip
from datetime import datetime
from itertools import repeat
from concurrent.futures import ProcessPoolExecutor
from geniepy.scripts.genieutils import decompress_gz
from geniepy.pubmed import PubMedArticle, ArticleSetParser


def is_xml_article_set(filename: str) -> bool:
    """
    Check if input file is pubmed xml compressed archive from name.

    Arguments:
        filename {str} -- the name of the file

    Returns:
        bool -- true if file is compressed pubmed article set
    """
    if filename.endswith(".xml.gz"):
        return True
    return False


def parse_pubmed_article_set(in_path: str, out_path: str):
    """
    Convert xml to json articles.

    Crawls through all files in a directory and create equivalent parsed jsonl file
    in output_path

    Arguments:
        in_path {str} -- absolute path to directory containing compressed article sets
        out_path {str} -- absolute path to desired directory to save output jsonl files
    """
    filename = os.path.basename(in_path)
    if not is_xml_article_set(filename):
        return
    xml_file = in_path.replace(".gz", "")
    if not os.path.exists(xml_file):
        logging.info("Extracting %s to %s", in_path, xml_file)
        decompress_gz(in_path, xml_file)

    logging.info("Parsing %s", xml_file)
    article_list: PubMedArticle = ArticleSetParser.extract_articles(xml_file)

    # Done with xml - delete to free up space
    os.remove(xml_file)

    output_file = os.path.join(out_path, filename.replace(".xml.gz", ".jsonl"))
    logging.info("Generating %s", output_file)
    ArticleSetParser.articles_to_jsonl(article_list, output_file)

    logging.info("Compressing file: %s", output_file)
    with open(output_file, "rb") as jsonl_data:
        data_jsonl = jsonl_data.read()
    compressed_data = gzip.compress(data_jsonl)
    output_file_compressed = output_file + ".gz"
    with open(output_file_compressed, "wb") as out_compressed:
        out_compressed.write(compressed_data)

    # Done with parsed file - delete to free up space
    os.remove(output_file)

    logging.info(
        "PID: %s. File Processed: %s. Articles Processed %s",
        os.getpid(),
        output_file,
        len(article_list),
    )
    return


def spawn_processes(data_in_dir: str, data_out_dir: str, max_workers: int):
    """
    Spawns processes to processes article sets in parallel.

    Creates process pool executor to parse article sets and generate output files.:w


    Arguments:
        data_in_dir {str} -- absolute path to input directory containing all articles
        data_out_dir {str} -- absolute path to output directory
        max_workers {int} -- max number of parallel processes to be created
    """
    start_time = datetime.now()
    xml_files: [str] = []
    for filename in os.listdir(data_in_dir):
        if is_xml_article_set(filename):
            xml_files.append(os.path.join(data_in_dir, filename))

    logging.info("Found %s PubMed article sets", len(xml_files))

    with ProcessPoolExecutor(max_workers=max_workers) as executor:
        executor.map(parse_pubmed_article_set, xml_files, repeat(data_out_dir))

    end_time = datetime.now()
    total_time = end_time - start_time
    logging.info(
        "Concurrent Process %s Execution: %s files, in %s",
        max_workers,
        len(xml_files),
        total_time,
    )


if __name__ == "__main__":

    logging.getLogger().setLevel(logging.INFO)

    ERROR_MSG = "Command line arguments expected: <path to data dir>, \
        <path to output dir>, <number of parallel processes (2-16)>"

    if not sys.argv or len(sys.argv) < 4:
        raise ValueError(ERROR_MSG)

    # check command line argument for input directory
    if os.path.isdir(sys.argv[1]):
        DATA_IN_DIR = sys.argv[1]
    else:
        raise ValueError("Input data directory is not valid. " + ERROR_MSG)

    # check command line argument for output directory
    if os.path.isdir(sys.argv[2]):
        DATA_OUT_DIR = sys.argv[2]
    else:
        raise ValueError("Output data directory is not valid. " + ERROR_MSG)

    # check argument for number of parallel processes
    try:
        MAX_WORKERS = int(sys.argv[3])
        if MAX_WORKERS < 1 or MAX_WORKERS > 16:
            logging.warning(
                "Invalid number of workers requested - Setting number of \
                workers to 1"
            )
            MAX_WORKERS = 1

        logging.info("Initializing parallel processing . . .")
        spawn_processes(DATA_IN_DIR, DATA_OUT_DIR, MAX_WORKERS)
    except ValueError:
        logging.error(
            "Max number of processes should be a valid integer. %s", ERROR_MSG
        )



"""
Module to extract citation metadata for a given range of PubMed articles.

This script depends upon Google cloud function to scale-up the extraction process.
Make sure the Google cloud functions are deployed and available before running
this script. (See: citation_cloud_function.py)
"""
from concurrent.futures import ProcessPoolExecutor
from datetime import datetime
from random import randint
import json
import os
import sys
import time
import requests

# Constants
_MIN_PUBMED_ID = 1
_MAX_PUBMED_ID = 40000000
_CHUNK_SIZE = 10
_CSV_SEPRATOR = "|"


# Google Cloud Functions - to scale-up data extraction
CLOUD_FN_1 = (
    "https://us-central1-csci-e-599-trend-detection.cloudfunctions.net/get_citation_1"
)
CLOUD_FN_2 = (
    "https://us-central1-csci-e-599-trend-detection.cloudfunctions.net/get_citation_2"
)
CLOUD_FN_3 = (
    "https://us-central1-csci-e-599-trend-detection.cloudfunctions.net/get_citation_3"
)
CLOUD_FN_4 = (
    "https://us-central1-csci-e-599-trend-detection.cloudfunctions.net/get_citation_4"
)
CLOUD_FN_5 = (
    "https://us-central1-csci-e-599-trend-detection.cloudfunctions.net/get_citation_5"
)
CLOUD_FN_6 = (
    "https://us-central1-csci-e-599-trend-detection.cloudfunctions.net/get_citation_6"
)
CLOUD_FN_7 = (
    "https://us-central1-csci-e-599-trend-detection.cloudfunctions.net/get_citation_7"
)
CLOUD_FN_8 = (
    "https://us-central1-csci-e-599-trend-detection.cloudfunctions.net/get_citation_8"
)
CLOUD_FN_9 = (
    "https://us-central1-csci-e-599-trend-detection.cloudfunctions.net/get_citation_9"
)
CLOUD_FN_10 = (
    "https://us-central1-csci-e-599-trend-detection.cloudfunctions.net/get_citation_10"
)

citation_dict = {}


def generate_ID_List(StartID, EndID):
    id_list = list(range(StartID, EndID + 1))
    chucked_id_list = [
        id_list[i * _CHUNK_SIZE : (i + 1) * _CHUNK_SIZE]  # noqa
        for i in range((len(id_list) + _CHUNK_SIZE - 1) // _CHUNK_SIZE)
    ]
    return chucked_id_list


def get_random_fn():
    # pick a random number between 1 and 5
    num = randint(1, 10)

    # return a randomly selected cloud function
    if num == 1:
        return CLOUD_FN_1
    elif num == 2:
        return CLOUD_FN_2
    elif num == 3:
        return CLOUD_FN_3
    elif num == 4:
        return CLOUD_FN_4
    elif num == 5:
        return CLOUD_FN_5
    elif num == 6:
        return CLOUD_FN_6
    elif num == 7:
        return CLOUD_FN_7
    elif num == 8:
        return CLOUD_FN_8
    elif num == 9:
        return CLOUD_FN_9
    else:
        return CLOUD_FN_10


def get_citation(IDList):
    ids_str = ",".join([str(elem) for elem in IDList])
    print(f"PID {os.getpid()}: Process Articles: {ids_str}")

    # construct api
    api_headers = {"content-type": "application/json"}
    api_body = {"pubmed_ids": ids_str}

    # call api - loop untill sucessful
    success = False
    while not success:
        cloud_fn = get_random_fn()
        try:
            req = requests.post(
                url=cloud_fn,
                data=json.dumps(api_body),
                headers=api_headers,
                timeout=(2, 5),
            )
            if req.status_code == 200:
                return req.text
        except Exception:
            """ if request fails or timeout, wait sometime and try again
            with another randomly selected cloud function """
            time.sleep(2)


def check_zero_citations(citationList):
    if not citationList or len(citationList) <= 0 or citationList == ["0"]:
        return True

    return False


def write_output(citation_dict, OutputDir):
    out_file = f"output-{datetime.now().microsecond}.csv"
    out_path = os.path.join(OutputDir, out_file)
    with open(out_path, "w") as file:
        file.write(
            "article_id"
            + _CSV_SEPRATOR
            + "cited_count"
            + _CSV_SEPRATOR
            + "cited_by_id"
            + "\n"
        )
        for key in citation_dict:
            if not check_zero_citations(citation_dict[key]):
                file.write(
                    key
                    + _CSV_SEPRATOR
                    + str(len(citation_dict[key]))
                    + _CSV_SEPRATOR
                    + ",".join(citation_dict[key])
                    + "\n"
                )
        print(
            f"Output: Citation data for {len(citation_dict)} \
PubMed Articles saved in file: {out_path}"
        )


def main(IDList, MaxWorkers, OutputDir):
    # start parallel processing
    with ProcessPoolExecutor(max_workers=MaxWorkers) as executor:
        result = executor.map(get_citation, IDList)

    # join parallel processing results
    for value in result:
        citation_dict.update(json.loads(value))

    # write json output
    write_output(citation_dict, OutputDir)


if __name__ == "__main__":  # noqa

    # command line arguments
    _start_id = 0
    _end_id = 0
    _max_workers = 0
    _out_dir = ""

    ERROR_MSG = """Command line arguments expected:
    <Start PubMed ID (ex: 1001)>,
    <End PubMed ID (ex: 1999)>,
    <Number of Parallel Processes (ex: 2-16)>
    <Output Directory (ex: ./data/)"""

    # check number if command line arguments
    if not sys.argv or len(sys.argv) < 5:
        raise ValueError(ERROR_MSG)

    # check command line argument for Start PubMed ID
    try:
        _start_id = int(sys.argv[1])
    except Exception:
        raise ValueError(f"Start PubMed ID is not valid. {ERROR_MSG}")

    # check command line argument for Start PubMed ID between allowed min and max values
    if _start_id < _MIN_PUBMED_ID or _start_id > (_MAX_PUBMED_ID - 1):
        raise ValueError(
            f"Start PubMed ID should be between {_MIN_PUBMED_ID} \
and {_MAX_PUBMED_ID-1}.  {ERROR_MSG}"
        )

    # check command line argument for End PubMed ID
    try:
        _end_id = int(sys.argv[2])
    except Exception:
        raise ValueError(f"End PubMed ID is not valid. {ERROR_MSG}")

    # check command line argument for End PubMed ID between allowed min and max values
    if _end_id < (_MIN_PUBMED_ID + 1) or _end_id > _MAX_PUBMED_ID:
        raise ValueError(
            f"End PubMed ID should be between {_MIN_PUBMED_ID+1} and {_MAX_PUBMED_ID}"
        )

    # check command line argument for Start PubMed ID < End PubMed ID
    if _start_id > _end_id:
        raise ValueError(
            f"End PubMed ID should be greater than Start PubMed ID. {ERROR_MSG}"
        )

    # check command line argument for Parallel Workers
    try:
        _max_workers = int(sys.argv[3])
    except Exception:
        raise ValueError(f"Number of parallel workers is not valid. {ERROR_MSG}")

    # check command line argument for Parallel Workers is between 2 and 16
    if _max_workers < 2 or _max_workers > 16:
        raise ValueError(
            f"Number of parallel workers should be between 2 and 16. {ERROR_MSG}"
        )

    # check command line argument for Output Directory is valid
    _out_dir = sys.argv[4]
    if not os.path.isdir(_out_dir):
        raise ValueError(f"Output data directory is not valid. {ERROR_MSG}")

    start_time = datetime.now()
    main(generate_ID_List(_start_id, _end_id), _max_workers, _out_dir)

    tot_time = datetime.now() - start_time
    print(
        f"Processed {_end_id-_start_id+1} PubMed Articles in \
{round(tot_time.total_seconds()/60,2)} minutes"
    )
